"""Tests for DLP scanning pipeline."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock

import pytest

from anteroom.config import DlpConfig, DlpPatternConfig
from anteroom.services.dlp import BUILTIN_PATTERNS, DlpScanner

# --- Helpers ---


def _make_config(**overrides: object) -> DlpConfig:
    defaults: dict[str, object] = {"enabled": True}
    defaults.update(overrides)
    return DlpConfig(**defaults)  # type: ignore[arg-type]


def _make_scanner(**overrides: object) -> DlpScanner:
    return DlpScanner(_make_config(**overrides))


# --- DlpConfig validation ---


class TestDlpConfig:
    def test_defaults(self) -> None:
        cfg = DlpConfig()
        assert cfg.enabled is False
        assert cfg.scan_output is True
        assert cfg.scan_input is False
        assert cfg.action == "redact"
        assert cfg.patterns == []
        assert cfg.custom_patterns == []
        assert cfg.redaction_string == "[REDACTED]"
        assert cfg.log_detections is True

    def test_invalid_action_defaults_to_redact(self) -> None:
        cfg = DlpConfig(action="invalid")
        assert cfg.action == "redact"

    def test_valid_actions_unchanged(self) -> None:
        for action in ("redact", "block", "warn"):
            cfg = DlpConfig(action=action)
            assert cfg.action == action


# --- Scanner disabled ---


class TestDlpScannerDisabled:
    def test_disabled_scanner_returns_pass(self) -> None:
        scanner = DlpScanner(DlpConfig(enabled=False))
        result = scanner.scan("123-45-6789")
        assert not result.matched
        assert result.action == "pass"

    def test_disabled_scanner_apply_returns_original(self) -> None:
        scanner = DlpScanner(DlpConfig(enabled=False))
        text, result = scanner.apply("123-45-6789")
        assert text == "123-45-6789"
        assert not result.matched


# --- Scanner properties ---


class TestDlpScannerProperties:
    def test_enabled_property(self) -> None:
        scanner = _make_scanner(enabled=True)
        assert scanner.enabled is True

    def test_scan_output_property(self) -> None:
        scanner = _make_scanner(scan_output=True)
        assert scanner.scan_output is True

    def test_scan_input_property(self) -> None:
        scanner = _make_scanner(scan_input=True)
        assert scanner.scan_input is True


# --- Built-in patterns ---


class TestBuiltInPatterns:
    def test_builtin_patterns_loaded_when_no_custom(self) -> None:
        scanner = _make_scanner()
        assert len(scanner._rules) == len(BUILTIN_PATTERNS)

    @pytest.mark.parametrize(
        "text",
        [
            "My SSN is 123-45-6789",
            "SSN: 999-88-7777",
            "Number 111-22-3333 is sensitive",
        ],
    )
    def test_ssn_detected(self, text: str) -> None:
        scanner = _make_scanner()
        result = scanner.scan(text)
        assert result.matched
        assert any(m.rule_name == "ssn" for m in result.matches)

    @pytest.mark.parametrize(
        "text",
        [
            "Phone: 123-456-7890",
            "Date: 2024-01-15",
            "Code: ABC-DE-FGHI",
        ],
    )
    def test_ssn_not_false_positive(self, text: str) -> None:
        scanner = _make_scanner(
            patterns=[DlpPatternConfig(name="ssn", pattern=r"\b\d{3}-\d{2}-\d{4}\b", description="SSN")]
        )
        result = scanner.scan(text)
        ssn_matches = [m for m in result.matches if m.rule_name == "ssn"]
        assert len(ssn_matches) == 0

    @pytest.mark.parametrize(
        "text",
        [
            "Card: 4111111111111111",
            "Visa: 4111 1111 1111 1111",
            "MC: 5500-0000-0000-0004",
        ],
    )
    def test_credit_card_detected(self, text: str) -> None:
        scanner = _make_scanner()
        result = scanner.scan(text)
        assert result.matched
        assert any(m.rule_name == "credit_card" for m in result.matches)

    @pytest.mark.parametrize(
        "text",
        [
            "Email: user@example.com",
            "Contact john.doe+tag@company.co.uk for info",
        ],
    )
    def test_email_detected(self, text: str) -> None:
        scanner = _make_scanner()
        result = scanner.scan(text)
        assert result.matched
        assert any(m.rule_name == "email" for m in result.matches)

    @pytest.mark.parametrize(
        "text",
        [
            "Call 555-123-4567",
            "Tel: +1-555-123-4567",
        ],
    )
    def test_phone_detected(self, text: str) -> None:
        scanner = _make_scanner()
        result = scanner.scan(text)
        assert result.matched
        assert any(m.rule_name == "phone_us" for m in result.matches)


# --- Custom patterns ---


class TestCustomPatterns:
    def test_custom_pattern_matches(self) -> None:
        scanner = _make_scanner(
            patterns=[DlpPatternConfig(name="acct", pattern=r"ACCT-\d{8}", description="Account number")]
        )
        result = scanner.scan("Your account ACCT-12345678 is active")
        assert result.matched
        assert result.matches[0].rule_name == "acct"
        assert result.matches[0].match_count == 1

    def test_custom_pattern_no_match(self) -> None:
        scanner = _make_scanner(
            patterns=[DlpPatternConfig(name="acct", pattern=r"ACCT-\d{8}", description="Account number")]
        )
        result = scanner.scan("No account numbers here")
        assert not result.matched

    def test_invalid_regex_skipped(self) -> None:
        scanner = _make_scanner(
            patterns=[
                DlpPatternConfig(name="bad", pattern="[invalid", description="broken"),
                DlpPatternConfig(name="good", pattern=r"\bsecret\b", description="keyword"),
            ]
        )
        assert len(scanner._rules) == 1
        assert scanner._rules[0].name == "good"

    def test_empty_name_or_pattern_skipped(self) -> None:
        scanner = _make_scanner(
            patterns=[
                DlpPatternConfig(name="", pattern=r"\d+", description="no name"),
                DlpPatternConfig(name="nopattern", pattern="", description="no pattern"),
                DlpPatternConfig(name="valid", pattern=r"\btest\b", description="valid"),
            ]
        )
        assert len(scanner._rules) == 1
        assert scanner._rules[0].name == "valid"

    def test_custom_patterns_merged_with_configured(self) -> None:
        cfg = DlpConfig(
            enabled=True,
            patterns=[DlpPatternConfig(name="p1", pattern=r"\bp1\b", description="first")],
            custom_patterns=[DlpPatternConfig(name="p2", pattern=r"\bp2\b", description="second")],
        )
        scanner = DlpScanner(cfg)
        assert len(scanner._rules) == 2
        names = {r.name for r in scanner._rules}
        assert names == {"p1", "p2"}


# --- Actions ---


class TestScanActions:
    def test_redact_replaces_matches(self) -> None:
        scanner = _make_scanner(action="redact")
        text, result = scanner.apply("SSN: 123-45-6789")
        assert result.matched
        assert result.action == "redact"
        assert "123-45-6789" not in text
        assert "[REDACTED]" in text

    def test_redact_custom_string(self) -> None:
        scanner = _make_scanner(action="redact", redaction_string="***")
        text, result = scanner.apply("SSN: 123-45-6789")
        assert "***" in text
        assert "123-45-6789" not in text

    def test_block_returns_empty(self) -> None:
        scanner = _make_scanner(action="block")
        text, result = scanner.apply("SSN: 123-45-6789")
        assert result.matched
        assert result.action == "block"
        assert text == ""

    def test_warn_returns_original(self) -> None:
        scanner = _make_scanner(action="warn")
        text, result = scanner.apply("SSN: 123-45-6789")
        assert result.matched
        assert result.action == "warn"
        assert text == "SSN: 123-45-6789"

    def test_no_match_returns_original(self) -> None:
        scanner = _make_scanner()
        text, result = scanner.apply("Hello world, nothing sensitive here")
        assert not result.matched
        assert text == "Hello world, nothing sensitive here"


# --- Direction filtering ---


class TestDirectionFiltering:
    def test_output_scan_skipped_when_disabled(self) -> None:
        scanner = _make_scanner(scan_output=False)
        result = scanner.scan("SSN: 123-45-6789", direction="output")
        assert not result.matched

    def test_input_scan_skipped_when_disabled(self) -> None:
        scanner = _make_scanner(scan_input=False)
        result = scanner.scan("SSN: 123-45-6789", direction="input")
        assert not result.matched

    def test_input_scan_works_when_enabled(self) -> None:
        scanner = _make_scanner(scan_input=True)
        result = scanner.scan("SSN: 123-45-6789", direction="input")
        assert result.matched

    def test_output_scan_works_when_enabled(self) -> None:
        scanner = _make_scanner(scan_output=True)
        result = scanner.scan("SSN: 123-45-6789", direction="output")
        assert result.matched


# --- Edge cases ---


class TestEdgeCases:
    def test_empty_string(self) -> None:
        scanner = _make_scanner()
        result = scanner.scan("")
        assert not result.matched

    def test_whitespace_only(self) -> None:
        scanner = _make_scanner()
        result = scanner.scan("   \n\t  ")
        assert not result.matched

    def test_multiple_matches_same_rule(self) -> None:
        scanner = _make_scanner(
            patterns=[DlpPatternConfig(name="ssn", pattern=r"\b\d{3}-\d{2}-\d{4}\b", description="SSN")]
        )
        result = scanner.scan("SSNs: 123-45-6789 and 987-65-4321")
        assert result.matched
        assert result.matches[0].match_count == 2

    def test_multiple_rules_match(self) -> None:
        scanner = _make_scanner()
        result = scanner.scan("SSN: 123-45-6789, email: user@example.com")
        assert result.matched
        assert len(result.matches) >= 2
        names = {m.rule_name for m in result.matches}
        assert "ssn" in names
        assert "email" in names

    def test_redact_multiple_patterns(self) -> None:
        scanner = _make_scanner(action="redact")
        text, result = scanner.apply("SSN: 123-45-6789, email: user@example.com")
        assert "123-45-6789" not in text
        assert "user@example.com" not in text
        assert text.count("[REDACTED]") >= 2


# --- Security: ReDoS protection and length guards ---


class TestSecurityGuards:
    def test_max_scan_length_truncates(self) -> None:
        from anteroom.services.dlp import MAX_SCAN_LENGTH

        scanner = _make_scanner(patterns=[DlpPatternConfig(name="test", pattern=r"SENSITIVE", description="test")])
        # Pattern placed beyond the max scan length — should NOT be detected
        text = "x" * (MAX_SCAN_LENGTH + 100) + "SENSITIVE"
        result = scanner.scan(text)
        assert not result.matched

    def test_within_scan_length_detected(self) -> None:
        scanner = _make_scanner(patterns=[DlpPatternConfig(name="test", pattern=r"SENSITIVE", description="test")])
        text = "SENSITIVE plus more text"
        result = scanner.scan(text)
        assert result.matched

    def test_pathological_pattern_rejected(self) -> None:
        scanner = _make_scanner(patterns=[DlpPatternConfig(name="evil", pattern=r"(a+)+$", description="ReDoS")])
        # The pathological pattern should be rejected during init
        assert len(scanner._rules) == 0

    def test_builtin_patterns_are_redos_safe(self) -> None:
        scanner = _make_scanner()
        # Ensure all built-in patterns load (none rejected by safety check)
        assert len(scanner._rules) == len(BUILTIN_PATTERNS)

    def test_iban_detected(self) -> None:
        scanner = _make_scanner()
        result = scanner.scan("IBAN: GB29NWBK60161331926819")
        assert result.matched
        assert any(m.rule_name == "iban" for m in result.matches)


# --- Agent loop integration ---


class TestAgentLoopIntegration:
    @pytest.mark.asyncio
    async def test_token_redacted_in_agent_loop(self) -> None:
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "Your SSN is 123-45-6789"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        scanner = _make_scanner(action="redact")

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=scanner,
        ):
            events.append(event)

        token_events = [e for e in events if e.kind == "token"]
        assert len(token_events) == 1
        assert "123-45-6789" not in token_events[0].data["content"]
        assert "[REDACTED]" in token_events[0].data["content"]

    @pytest.mark.asyncio
    async def test_block_emits_dlp_blocked_event(self) -> None:
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "SSN: 123-45-6789"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        scanner = _make_scanner(action="block")

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=scanner,
        ):
            events.append(event)

        blocked_events = [e for e in events if e.kind == "dlp_blocked"]
        assert len(blocked_events) == 1
        assert blocked_events[0].data["direction"] == "output"
        # Block should NOT emit a done event after blocking
        done_events = [e for e in events if e.kind == "done"]
        assert len(done_events) == 0

    @pytest.mark.asyncio
    async def test_warn_emits_dlp_warning_event(self) -> None:
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "SSN: 123-45-6789"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        scanner = _make_scanner(action="warn")

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=scanner,
        ):
            events.append(event)

        warning_events = [e for e in events if e.kind == "dlp_warning"]
        # Exactly 1 warning: from the final assembled-text scan only (no per-chunk duplicate)
        assert len(warning_events) == 1

        # Token content should be preserved (warn doesn't modify)
        token_events = [e for e in events if e.kind == "token"]
        assert any("123-45-6789" in e.data["content"] for e in token_events)

    @pytest.mark.asyncio
    async def test_no_scanner_passes_through(self) -> None:
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "SSN: 123-45-6789"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=None,
        ):
            events.append(event)

        token_events = [e for e in events if e.kind == "token"]
        assert "123-45-6789" in token_events[0].data["content"]

    @pytest.mark.asyncio
    async def test_final_scan_catches_cross_chunk_pattern(self) -> None:
        """Final assembled-text scan catches patterns split across streaming chunks."""
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        # SSN split across two chunks: "123-45" + "-6789"
        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "SSN is 123-45"}}
            yield {"event": "token", "data": {"content": "-6789 ok"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        scanner = _make_scanner(action="block")

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=scanner,
        ):
            events.append(event)

        # The final assembled-text scan should catch the cross-chunk SSN
        blocked = [e for e in events if e.kind == "dlp_blocked"]
        assert len(blocked) == 1

    @pytest.mark.asyncio
    async def test_scan_output_false_bypasses_dlp_in_agent_loop(self) -> None:
        """Scanner with scan_output=False should not scan agent output."""
        from anteroom.services.agent_loop import AgentEvent, run_agent_loop
        from anteroom.services.ai_service import AIService

        ai_service = AIService.__new__(AIService)
        ai_service.config = MagicMock()
        ai_service.config.narration_cadence = 0
        ai_service._token_provider = None
        ai_service.client = MagicMock()

        async def fake_stream_chat(messages, **kwargs):
            yield {"event": "token", "data": {"content": "SSN: 123-45-6789"}}
            yield {"event": "done", "data": {}}

        ai_service.stream_chat = fake_stream_chat

        scanner = _make_scanner(action="block", scan_output=False)

        events: list[AgentEvent] = []
        async for event in run_agent_loop(
            ai_service=ai_service,
            messages=[{"role": "user", "content": "test"}],
            tool_executor=AsyncMock(),
            tools_openai=None,
            dlp_scanner=scanner,
        ):
            events.append(event)

        # No blocking — scan_output is disabled
        blocked = [e for e in events if e.kind == "dlp_blocked"]
        assert len(blocked) == 0
        # Token passes through unmodified
        tokens = [e for e in events if e.kind == "token"]
        assert "123-45-6789" in tokens[0].data["content"]


# --- Config parsing ---


class TestDlpConfigParsing:
    def test_default_dlp_in_safety_config(self) -> None:
        from anteroom.config import SafetyConfig

        cfg = SafetyConfig()
        assert isinstance(cfg.dlp, DlpConfig)
        assert cfg.dlp.enabled is False

    def test_dlp_from_yaml(self, tmp_path) -> None:
        import yaml

        from anteroom.config import load_config

        config_data = {
            "ai": {"api_key": "test-key", "model": "gpt-4", "base_url": "http://localhost:1234"},
            "safety": {
                "dlp": {
                    "enabled": True,
                    "action": "block",
                    "scan_input": True,
                    "patterns": [
                        {"name": "ssn", "pattern": r"\b\d{3}-\d{2}-\d{4}\b", "description": "SSN"},
                    ],
                    "custom_patterns": [
                        {"name": "acct", "pattern": r"ACCT-\d+", "description": "Account"},
                    ],
                }
            },
        }
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump(config_data))
        cfg, _ = load_config(cfg_file)
        assert cfg.safety.dlp.enabled is True
        assert cfg.safety.dlp.action == "block"
        assert cfg.safety.dlp.scan_input is True
        assert len(cfg.safety.dlp.patterns) == 1
        assert cfg.safety.dlp.patterns[0].name == "ssn"
        assert len(cfg.safety.dlp.custom_patterns) == 1
        assert cfg.safety.dlp.custom_patterns[0].name == "acct"

    def test_dlp_env_var_override(self, tmp_path, monkeypatch) -> None:
        import yaml

        from anteroom.config import load_config

        monkeypatch.setenv("AI_CHAT_DLP_ENABLED", "true")
        monkeypatch.setenv("AI_CHAT_DLP_ACTION", "warn")
        config_data = {"ai": {"api_key": "test-key", "model": "gpt-4", "base_url": "http://localhost:1234"}}
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump(config_data))
        cfg, _ = load_config(cfg_file)
        assert cfg.safety.dlp.enabled is True
        assert cfg.safety.dlp.action == "warn"

    def test_dlp_invalid_action_from_yaml(self, tmp_path) -> None:
        import yaml

        from anteroom.config import load_config

        config_data = {
            "ai": {"api_key": "test-key", "model": "gpt-4", "base_url": "http://localhost:1234"},
            "safety": {"dlp": {"enabled": True, "action": "invalid_action"}},
        }
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump(config_data))
        cfg, _ = load_config(cfg_file)
        assert cfg.safety.dlp.action == "redact"

    def test_dlp_malformed_patterns_skipped(self, tmp_path) -> None:
        import yaml

        from anteroom.config import load_config

        config_data = {
            "ai": {"api_key": "test-key", "model": "gpt-4", "base_url": "http://localhost:1234"},
            "safety": {
                "dlp": {
                    "enabled": True,
                    "patterns": [
                        {"name": "valid", "pattern": r"\d+", "description": "numbers"},
                        {"name": "", "pattern": r"\d+"},  # empty name
                        {"pattern": r"\d+"},  # missing name
                        "not_a_dict",  # wrong type
                    ],
                }
            },
        }
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump(config_data))
        cfg, _ = load_config(cfg_file)
        assert len(cfg.safety.dlp.patterns) == 1
        assert cfg.safety.dlp.patterns[0].name == "valid"

    def test_audit_events_include_dlp_by_default(self) -> None:
        from anteroom.config import AuditConfig

        cfg = AuditConfig()
        assert cfg.events.get("dlp") is True



"""Tests for on_failure branches (#1127) and retry_on_result (#1129)."""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, patch

import pytest

from anteroom.config import WorkflowConfig
from anteroom.db import init_db
from anteroom.services.workflow_engine import (
    WorkflowEngine,
    load_definition,
    register_gate_condition,
)
from anteroom.services.workflow_runners import RunnerResult, create_default_registry
from anteroom.services.workflow_storage import list_workflow_events, list_workflow_steps


@pytest.fixture()
def db():
    with tempfile.TemporaryDirectory() as td:
        conn = init_db(Path(td) / "test.db")
        yield conn
        conn.close()


@pytest.fixture()
def engine(db: Any) -> WorkflowEngine:
    config = WorkflowConfig()
    registry = create_default_registry()
    return WorkflowEngine(db, config, registry)


@pytest.fixture(autouse=True)
def _register_test_gates():
    async def always_pass(run: Any, step: Any, inputs: Any) -> bool:
        return True

    register_gate_condition("always_pass", always_pass)
    yield


# ---------------------------------------------------------------------------
# YAML fixtures
# ---------------------------------------------------------------------------

ON_FAILURE_BASIC = """\
kind: workflow
id: on_failure_test
version: 0.1.0
inputs: {}
steps:
  - id: run_tests
    type: runner
    runner: shell
    command: "echo testing"
    on_failure:
      - id: collect_logs
        type: runner
        runner: shell
        command: "echo collecting logs"
      - id: notify
        type: runner
        runner: shell
        command: "echo notifying"
  - id: deploy
    type: runner
    runner: shell
    command: "echo deploying"
"""

ON_FAILURE_BRANCH_FAILS = """\
kind: workflow
id: on_failure_branch_fail
version: 0.1.0
inputs: {}
steps:
  - id: run_tests
    type: runner
    runner: shell
    command: "echo testing"
    on_failure:
      - id: bad_handler
        type: runner
        runner: shell
        command: "echo handler fails"
  - id: deploy
    type: runner
    runner: shell
    command: "echo deploying"
"""

ON_FAILURE_WITH_CONTINUE = """\
kind: workflow
id: on_failure_continue
version: 0.1.0
inputs: {}
steps:
  - id: run_tests
    type: runner
    runner: shell
    command: "echo testing"
    on_failure:
      - id: collect_logs
        type: runner
        runner: shell
        command: "echo logs"
    continue_on: [failed]
  - id: deploy
    type: runner
    runner: shell
    command: "echo deploying"
"""

RETRY_ON_RESULT = """\
kind: workflow
id: retry_on_result_test
version: 0.1.0
inputs: {}
steps:
  - id: generate
    type: runner
    runner: shell
    command: "echo generating"
    retry:
      max_attempts: 3
      backoff: fixed
      initial_delay: 0.01
      retry_on_result:
        - field: result_artifacts.exit_code
          not_equals: "0"
"""


# ---------------------------------------------------------------------------
# Tests: on_failure (#1127)
# ---------------------------------------------------------------------------


class TestOnFailureBranch:
    @pytest.mark.asyncio
    async def test_branch_runs_on_failure_and_continues(self, db: Any, engine: WorkflowEngine) -> None:
        """Step fails, on_failure branch succeeds → run continues to next step."""
        call_count = 0

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            nonlocal call_count
            call_count += 1
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="tests failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BASIC)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        steps = list_workflow_steps(db, run["id"])
        step_ids = [s["step_id"] for s in steps]
        assert "run_tests/_fail/collect_logs" in step_ids
        assert "run_tests/_fail/notify" in step_ids
        assert "deploy" in step_ids

    @pytest.mark.asyncio
    async def test_branch_fails_aborts_run(self, db: Any, engine: WorkflowEngine) -> None:
        """Step fails, on_failure branch also fails → run aborts."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="tests failed")
            if step_id == "bad_handler":
                return RunnerResult(status="failed", summary="handler failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BRANCH_FAILS)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "failed"
        steps = list_workflow_steps(db, run["id"])
        step_ids = [s["step_id"] for s in steps]
        assert "run_tests/_fail/bad_handler" in step_ids
        assert "deploy" not in step_ids

    @pytest.mark.asyncio
    async def test_branch_not_triggered_on_success(self, db: Any, engine: WorkflowEngine) -> None:
        """Step succeeds → on_failure branch does not run."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BASIC)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        steps = list_workflow_steps(db, run["id"])
        step_ids = [s["step_id"] for s in steps]
        assert not any("_fail/" in sid for sid in step_ids)

    @pytest.mark.asyncio
    async def test_branch_step_ids_prefixed(self, db: Any, engine: WorkflowEngine) -> None:
        """Branch step IDs use parent/_fail/child convention."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BASIC)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        steps = list_workflow_steps(db, run["id"])
        fail_steps = [s for s in steps if "_fail/" in s["step_id"]]
        assert len(fail_steps) == 2
        for s in fail_steps:
            assert s["step_id"].startswith("run_tests/_fail/")

    @pytest.mark.asyncio
    async def test_failure_branch_events_emitted(self, db: Any, engine: WorkflowEngine) -> None:
        """failure_branch_started and failure_branch_completed events emitted."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BASIC)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        events = list_workflow_events(db, run["id"])
        event_types = [e["event_type"] for e in events]
        assert "failure_branch_started" in event_types
        assert "failure_branch_completed" in event_types

    @pytest.mark.asyncio
    async def test_on_failure_with_continue_on(self, db: Any, engine: WorkflowEngine) -> None:
        """on_failure + continue_on: branch runs, then continue_on allows progression."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_WITH_CONTINUE)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        steps = list_workflow_steps(db, run["id"])
        step_ids = [s["step_id"] for s in steps]
        assert "run_tests/_fail/collect_logs" in step_ids
        assert "deploy" in step_ids

    @pytest.mark.asyncio
    async def test_on_failure_with_retry_runs_after_exhaustion(self, db: Any, engine: WorkflowEngine) -> None:
        """retry + on_failure: retry first, branch after exhaustion."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="failed")
            return RunnerResult(status="success", summary="ok")

        yaml = """\
kind: workflow
id: retry_then_on_failure
version: 0.1.0
inputs: {}
steps:
  - id: run_tests
    type: runner
    runner: shell
    command: "echo test"
    retry:
      max_attempts: 2
      backoff: fixed
      initial_delay: 0.01
    on_failure:
      - id: handler
        type: runner
        runner: shell
        command: "echo handle"
  - id: next
    type: runner
    runner: shell
    command: "echo next"
"""
        defn = load_definition(yaml)
        with (
            patch(
                "anteroom.services.workflow_runners.execute_opaque_runner",
                side_effect=mock_runner,
            ),
            patch.object(WorkflowEngine, "_retry_delay", new_callable=AsyncMock),
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        steps = list_workflow_steps(db, run["id"])
        step_ids = [s["step_id"] for s in steps]
        assert "run_tests/_fail/handler" in step_ids
        assert "next" in step_ids
        retry_events = [e for e in list_workflow_events(db, run["id"]) if e["event_type"] == "step_retry"]
        assert len(retry_events) >= 1

    @pytest.mark.asyncio
    async def test_branch_accesses_parent_via_context_from(self, db: Any, engine: WorkflowEngine) -> None:
        """Branch step can reference parent's failure via step_results."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            step_id = kwargs.get("step_id", "?")
            if step_id == "run_tests":
                return RunnerResult(status="failed", summary="3 tests failed")
            return RunnerResult(status="success", summary="ok")

        defn = load_definition(ON_FAILURE_BASIC)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"


# ---------------------------------------------------------------------------
# Tests: on_failure validation
# ---------------------------------------------------------------------------


class TestOnFailureValidation:
    def test_nested_on_failure_rejected(self) -> None:
        yaml = """\
kind: workflow
id: bad
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: "echo x"
    on_failure:
      - id: handler
        type: runner
        runner: shell
        command: "echo y"
        on_failure:
          - id: nested
            type: runner
            runner: shell
            command: "echo z"
"""
        with pytest.raises(ValueError, match="nested on_failure"):
            load_definition(yaml)

    def test_gate_in_on_failure_rejected(self) -> None:
        yaml = """\
kind: workflow
id: bad
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: "echo x"
    on_failure:
      - id: bad_gate
        type: gate
        condition: always_pass
"""
        with pytest.raises(ValueError, match="'runner' or 'llm'"):
            load_definition(yaml)

    def test_on_failure_on_gate_step_rejected(self) -> None:
        yaml = """\
kind: workflow
id: bad
version: 0.1.0
inputs: {}
steps:
  - id: g1
    type: gate
    condition: always_pass
    on_failure:
      - id: handler
        type: runner
        runner: shell
        command: "echo y"
"""
        with pytest.raises(ValueError, match="not allowed on gate"):
            load_definition(yaml)

    def test_valid_on_failure_accepted(self) -> None:
        defn = load_definition(ON_FAILURE_BASIC)
        assert defn.steps[0].on_failure is not None
        assert len(defn.steps[0].on_failure) == 2


# ---------------------------------------------------------------------------
# Tests: retry_on_result (#1129)
# ---------------------------------------------------------------------------


class TestRetryOnResult:
    @pytest.mark.asyncio
    async def test_result_condition_triggers_retry(self, db: Any, engine: WorkflowEngine) -> None:
        """retry_on_result condition matches → retry triggered."""
        call_count = 0

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            nonlocal call_count
            call_count += 1
            if call_count <= 2:
                return RunnerResult(status="success", summary="bad", artifacts={"exit_code": "1"})
            return RunnerResult(status="success", summary="good", artifacts={"exit_code": "0"})

        defn = load_definition(RETRY_ON_RESULT)
        with (
            patch(
                "anteroom.services.workflow_runners.execute_opaque_runner",
                side_effect=mock_runner,
            ),
            patch.object(WorkflowEngine, "_retry_delay", new_callable=AsyncMock),
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        assert call_count == 3

    @pytest.mark.asyncio
    async def test_result_condition_no_match_accepts(self, db: Any, engine: WorkflowEngine) -> None:
        """retry_on_result condition doesn't match → result accepted."""

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            return RunnerResult(status="success", summary="good", artifacts={"exit_code": "0"})

        defn = load_definition(RETRY_ON_RESULT)
        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            side_effect=mock_runner,
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"

    @pytest.mark.asyncio
    async def test_retry_reason_in_event_payload(self, db: Any, engine: WorkflowEngine) -> None:
        """step_retry event includes retry_reason: result_condition."""
        call_count = 0

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return RunnerResult(status="success", summary="bad", artifacts={"exit_code": "1"})
            return RunnerResult(status="success", summary="good", artifacts={"exit_code": "0"})

        defn = load_definition(RETRY_ON_RESULT)
        with (
            patch(
                "anteroom.services.workflow_runners.execute_opaque_runner",
                side_effect=mock_runner,
            ),
            patch.object(WorkflowEngine, "_retry_delay", new_callable=AsyncMock),
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        events = list_workflow_events(db, run["id"])
        retry_events = [e for e in events if e["event_type"] == "step_retry"]
        assert len(retry_events) >= 1
        payload = retry_events[0].get("payload") or {}
        if isinstance(payload, str):
            import json

            payload = json.loads(payload)
        assert payload.get("retry_reason") == "result_condition"

    @pytest.mark.asyncio
    async def test_failed_status_retry_has_reason(self, db: Any, engine: WorkflowEngine) -> None:
        """Existing failed-status retry includes retry_reason: failed_status."""
        call_count = 0

        async def mock_runner(*args: Any, **kwargs: Any) -> RunnerResult:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return RunnerResult(status="failed", summary="bad")
            return RunnerResult(status="success", summary="good")

        yaml = """\
kind: workflow
id: failed_retry
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: "echo x"
    retry:
      max_attempts: 2
      backoff: fixed
      initial_delay: 0.01
"""
        defn = load_definition(yaml)
        with (
            patch(
                "anteroom.services.workflow_runners.execute_opaque_runner",
                side_effect=mock_runner,
            ),
            patch.object(WorkflowEngine, "_retry_delay", new_callable=AsyncMock),
        ):
            run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        assert run["status"] == "completed"
        events = list_workflow_events(db, run["id"])
        retry_events = [e for e in events if e["event_type"] == "step_retry"]
        assert len(retry_events) >= 1
        payload = retry_events[0].get("payload") or {}
        if isinstance(payload, str):
            import json

            payload = json.loads(payload)
        assert payload.get("retry_reason") == "failed_status"


class TestRetryOnResultValidation:
    def test_retry_on_result_missing_field_rejected(self) -> None:
        yaml = """\
kind: workflow
id: bad
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: "echo x"
    retry:
      max_attempts: 2
      retry_on_result:
        - equals: "0"
"""
        with pytest.raises(ValueError, match="missing 'field'"):
            load_definition(yaml)

    def test_retry_on_result_invalid_operator_rejected(self) -> None:
        yaml = """\
kind: workflow
id: bad
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: "echo x"
    retry:
      max_attempts: 2
      retry_on_result:
        - field: result_status
          greater_than: 5
"""
        with pytest.raises(ValueError, match="operator"):
            load_definition(yaml)

    def test_valid_retry_on_result_accepted(self) -> None:
        defn = load_definition(RETRY_ON_RESULT)
        assert defn.steps[0].retry is not None
        assert defn.steps[0].retry.get("retry_on_result") is not None



"""Tests for structured workflow step outputs (#1131)."""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any

import pytest

from anteroom.config import WorkflowConfig
from anteroom.db import init_db
from anteroom.services.workflow_engine import (
    WorkflowEngine,
    _resolve_step_outputs,
    load_definition,
    resolve_context_from,
)
from anteroom.services.workflow_runners import RunnerResult, create_default_registry
from anteroom.services.workflow_storage import (
    create_workflow_run,
    create_workflow_step,
    get_workflow_step,
    list_workflow_steps,
    update_workflow_step,
)


@pytest.fixture()
def db():
    with tempfile.TemporaryDirectory() as td:
        conn = init_db(Path(td) / "test.db")
        yield conn
        conn.close()


def _make_run(db: Any, **overrides: Any) -> dict[str, Any]:
    defaults = {
        "workflow_id": "test_outputs",
        "workflow_version": "0.1.0",
        "target_kind": "task",
        "target_ref": "1",
    }
    defaults.update(overrides)
    return create_workflow_run(db, **defaults)


# ---------------------------------------------------------------------------
# _resolve_step_outputs (unit)
# ---------------------------------------------------------------------------


class TestResolveStepOutputs:
    def test_resolve_from_artifacts(self) -> None:
        decls = [{"name": "exit_code", "from": "result_artifacts.exit_code"}]
        step_data = {"result_artifacts": {"exit_code": 0}}
        result = _resolve_step_outputs(decls, step_data, {})
        assert result == {"exit_code": 0}

    def test_resolve_from_status(self) -> None:
        decls = [{"name": "status", "from": "result_status"}]
        step_data = {"result_status": "success"}
        result = _resolve_step_outputs(decls, step_data, {})
        assert result == {"status": "success"}

    def test_runner_outputs_take_precedence(self) -> None:
        decls = [{"name": "count", "from": "result_artifacts.count"}]
        step_data = {"result_artifacts": {"count": 10}}
        runner_outputs = {"count": 42}
        result = _resolve_step_outputs(decls, step_data, runner_outputs)
        assert result == {"count": 42}

    def test_missing_path_resolves_to_none(self) -> None:
        decls = [{"name": "missing", "from": "result_artifacts.nonexistent"}]
        step_data = {"result_artifacts": {}}
        result = _resolve_step_outputs(decls, step_data, {})
        assert result == {"missing": None}

    def test_multiple_outputs(self) -> None:
        decls = [
            {"name": "code", "from": "result_artifacts.exit_code"},
            {"name": "model", "from": "result_artifacts.model"},
        ]
        step_data = {"result_artifacts": {"exit_code": 0, "model": "gpt-4"}}
        result = _resolve_step_outputs(decls, step_data, {})
        assert result == {"code": 0, "model": "gpt-4"}

    def test_deeply_nested_path(self) -> None:
        decls = [{"name": "val", "from": "result_artifacts.structured_output.data.key"}]
        step_data = {"result_artifacts": {"structured_output": {"data": {"key": "found"}}}}
        result = _resolve_step_outputs(decls, step_data, {})
        assert result == {"val": "found"}


# ---------------------------------------------------------------------------
# RunnerResult.outputs field
# ---------------------------------------------------------------------------


class TestRunnerResultOutputs:
    def test_default_empty(self) -> None:
        r = RunnerResult(status="success")
        assert r.outputs == {}

    def test_outputs_in_to_dict(self) -> None:
        r = RunnerResult(status="success", outputs={"key": "val"})
        d = r.to_dict()
        assert d["outputs"] == {"key": "val"}


# ---------------------------------------------------------------------------
# Validation
# ---------------------------------------------------------------------------


class TestOutputsValidation:
    def test_valid_outputs_declaration(self) -> None:
        yaml_str = """\
kind: workflow
id: test_outputs
version: 0.1.0
inputs: {}
steps:
  - id: run_tests
    type: runner
    runner: shell
    command: echo ok
    outputs:
      - name: exit_code
        from: result_artifacts.exit_code
"""
        defn = load_definition(yaml_str)
        assert defn.steps[0].outputs is not None
        assert len(defn.steps[0].outputs) == 1

    def test_output_missing_name_raises(self) -> None:
        yaml_str = """\
kind: workflow
id: test_outputs
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: echo ok
    outputs:
      - from: result_artifacts.exit_code
"""
        with pytest.raises(ValueError, match="missing 'name'"):
            load_definition(yaml_str)

    def test_output_missing_from_raises(self) -> None:
        yaml_str = """\
kind: workflow
id: test_outputs
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: echo ok
    outputs:
      - name: exit_code
"""
        with pytest.raises(ValueError, match="missing 'from'"):
            load_definition(yaml_str)

    def test_duplicate_output_name_raises(self) -> None:
        yaml_str = """\
kind: workflow
id: test_outputs
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: echo ok
    outputs:
      - name: dup
        from: result_status
      - name: dup
        from: result_summary
"""
        with pytest.raises(ValueError, match="duplicate output name"):
            load_definition(yaml_str)


# ---------------------------------------------------------------------------
# When clause with result_outputs
# ---------------------------------------------------------------------------


class TestWhenWithOutputs:
    def test_when_equals_on_result_outputs(self) -> None:
        when = {"step": "s1", "field": "result_outputs.exit_code", "equals": "0"}
        results = {"s1": {"result_outputs": {"exit_code": "0"}}}
        assert WorkflowEngine._evaluate_when(when, results) is True

    def test_when_not_equals_on_result_outputs(self) -> None:
        when = {"step": "s1", "field": "result_outputs.exit_code", "not_equals": "0"}
        results = {"s1": {"result_outputs": {"exit_code": "1"}}}
        assert WorkflowEngine._evaluate_when(when, results) is True

    def test_when_is_not_empty_on_result_outputs(self) -> None:
        when = {"step": "s1", "field": "result_outputs.report_path", "is_not_empty": True}
        results = {"s1": {"result_outputs": {"report_path": "/tmp/report.xml"}}}
        assert WorkflowEngine._evaluate_when(when, results) is True

    def test_when_field_validation_allows_result_outputs(self) -> None:
        yaml_str = """\
kind: workflow
id: test_outputs
version: 0.1.0
inputs: {}
steps:
  - id: s1
    type: runner
    runner: shell
    command: echo ok
  - id: s2
    type: runner
    runner: shell
    command: echo ok
    when:
      step: s1
      field: result_outputs.exit_code
      equals: "0"
"""
        defn = load_definition(yaml_str)
        assert defn.steps[1].when is not None


# ---------------------------------------------------------------------------
# Context from result_outputs
# ---------------------------------------------------------------------------


class TestContextFromOutputs:
    def test_resolve_context_from_result_outputs(self) -> None:
        refs = [{"step": "s1", "field": "result_outputs.report_path"}]
        results = {"s1": {"result_outputs": {"report_path": "/tmp/report.xml"}}}
        ctx = resolve_context_from(refs, results)
        assert "/tmp/report.xml" in ctx

    def test_resolve_context_from_missing_output(self) -> None:
        refs = [{"step": "s1", "field": "result_outputs.missing"}]
        results = {"s1": {"result_outputs": {}}}
        ctx = resolve_context_from(refs, results)
        assert ctx == ""


# ---------------------------------------------------------------------------
# Storage round-trip
# ---------------------------------------------------------------------------


class TestOutputsStorage:
    def test_update_and_get_step_outputs(self, db: Any) -> None:
        run = _make_run(db)
        step = create_workflow_step(db, run_id=run["id"], step_id="s1", step_type="runner", runner_type="shell")
        outputs = {"exit_code": 0, "report_path": "/tmp/report.xml"}
        update_workflow_step(db, step["id"], result_outputs=outputs)
        fetched = get_workflow_step(db, step["id"])
        assert fetched is not None
        assert fetched["result_outputs"] == outputs

    def test_outputs_none_by_default(self, db: Any) -> None:
        run = _make_run(db)
        step = create_workflow_step(db, run_id=run["id"], step_id="s1", step_type="runner")
        fetched = get_workflow_step(db, step["id"])
        assert fetched is not None
        assert fetched["result_outputs"] is None

    def test_outputs_survive_list_steps(self, db: Any) -> None:
        run = _make_run(db)
        create_workflow_step(db, run_id=run["id"], step_id="s1", step_type="runner", runner_type="shell")
        steps = list_workflow_steps(db, run["id"])
        step = steps[0]
        update_workflow_step(db, step["id"], result_outputs={"key": "val"})
        steps = list_workflow_steps(db, run["id"])
        assert steps[0]["result_outputs"] == {"key": "val"}


# ---------------------------------------------------------------------------
# Sanitize checkpoint data includes result_outputs
# ---------------------------------------------------------------------------


class TestSanitizeWithOutputs:
    def test_sanitize_preserves_result_outputs(self) -> None:
        data = {
            "s1": {
                "result_status": "success",
                "result_summary": "ok",
                "result_artifacts": {},
                "result_findings": None,
                "result_outputs": {"exit_code": 0, "report": "/tmp/r.xml"},
            }
        }
        result = WorkflowEngine._sanitize_checkpoint_data(data)
        assert result["s1"]["result_outputs"] == {"exit_code": 0, "report": "/tmp/r.xml"}

    def test_sanitize_handles_none_outputs(self) -> None:
        data = {
            "s1": {
                "result_status": "success",
                "result_summary": "ok",
                "result_artifacts": {},
                "result_findings": None,
                "result_outputs": None,
            }
        }
        result = WorkflowEngine._sanitize_checkpoint_data(data)
        assert result["s1"]["result_outputs"] is None

    def test_sanitize_handles_missing_outputs(self) -> None:
        data = {
            "s1": {
                "result_status": "success",
                "result_summary": "ok",
                "result_artifacts": {},
                "result_findings": None,
            }
        }
        result = WorkflowEngine._sanitize_checkpoint_data(data)
        assert result["s1"]["result_outputs"] is None


# ---------------------------------------------------------------------------
# Rebuild step results includes result_outputs
# ---------------------------------------------------------------------------


class TestRebuildWithOutputs:
    def test_rebuild_includes_outputs(self, db: Any) -> None:
        config = WorkflowConfig()
        registry = create_default_registry()
        engine = WorkflowEngine(db, config, registry)

        run = _make_run(db)
        step = create_workflow_step(db, run_id=run["id"], step_id="s1", step_type="runner", runner_type="shell")
        update_workflow_step(
            db,
            step["id"],
            status="completed",
            result_status="success",
            result_outputs={"exit_code": 0},
        )
        results = engine._rebuild_step_results(run["id"])
        assert results["s1"]["result_outputs"] == {"exit_code": 0}

    def test_rebuild_none_when_no_outputs(self, db: Any) -> None:
        config = WorkflowConfig()
        registry = create_default_registry()
        engine = WorkflowEngine(db, config, registry)

        run = _make_run(db)
        step = create_workflow_step(db, run_id=run["id"], step_id="s1", step_type="runner", runner_type="shell")
        update_workflow_step(
            db,
            step["id"],
            status="completed",
            result_status="success",
        )
        results = engine._rebuild_step_results(run["id"])
        assert results["s1"]["result_outputs"] is None



"""Snapshot-style tests for theme-aware renderer output.

Verifies that:
- set_theme() updates the module-level _theme and refreshes aliases
- _refresh_aliases() correctly updates module aliases from the theme
- Theme colors flow through render_error(), render_warning(), and
  _build_thinking_text() instead of being hardcoded
- All built-in themes round-trip correctly through set_theme()
"""

from __future__ import annotations

import io

import pytest
from rich.console import Console

import anteroom.cli.renderer as renderer
from anteroom.cli.renderer import (
    BusyStatus,
    _build_thinking_text,
    _refresh_aliases,
    format_status_toolbar,
    render_error,
    render_warning,
    set_theme,
)
from anteroom.cli.themes import _BUILTIN_THEMES, CliTheme

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _capture_console() -> tuple[Console, io.StringIO]:
    """Return a (console, buffer) pair that captures Rich output as plain text."""
    buf = io.StringIO()
    con = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
    return con, buf


def _plain(text: str) -> str:
    """Strip ANSI escapes for assertion-friendly comparisons."""
    import re

    return re.sub(r"\x1b\[[0-9;]*m", "", text)


def _hex_to_ansi_rgb(hex_color: str) -> str:
    """Convert #RRGGBB to the ANSI 24-bit foreground RGB sequence Rich emits."""
    h = hex_color.lstrip("#")
    r, g, b = int(h[0:2], 16), int(h[2:4], 16), int(h[4:6], 16)
    return f"38;2;{r};{g};{b}"


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture(autouse=True)
def _clear_no_color(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("NO_COLOR", raising=False)


@pytest.fixture(autouse=True)
def restore_midnight_theme() -> None:  # type: ignore[return]
    """Reset to midnight after every test so state doesn't leak."""
    yield
    set_theme(CliTheme.load("midnight"))


# ---------------------------------------------------------------------------
# set_theme() and _refresh_aliases()
# ---------------------------------------------------------------------------


class TestSetTheme:
    def test_set_theme_updates_module_theme(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer._theme is dawn

    def test_set_theme_refreshes_gold_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.GOLD == dawn.accent

    def test_set_theme_refreshes_muted_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.MUTED == dawn.muted

    def test_set_theme_refreshes_chrome_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.CHROME == dawn.chrome

    def test_set_theme_refreshes_error_red_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.ERROR_RED == dawn.error

    def test_set_theme_refreshes_slate_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.SLATE == dawn.secondary

    def test_set_theme_refreshes_blue_alias(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer.BLUE == dawn.logo_blue

    def test_set_theme_midnight_accent(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)
        assert renderer._theme.accent == midnight.accent

    def test_set_theme_dawn_accent(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer._theme.accent == dawn.accent

    def test_aliases_differ_between_themes(self) -> None:
        set_theme(CliTheme.load("midnight"))
        midnight_gold = renderer.GOLD

        set_theme(CliTheme.load("dawn"))
        dawn_gold = renderer.GOLD

        # Midnight accent is #C5A059; dawn accent is #B8860B — they must differ
        assert midnight_gold != dawn_gold

    def test_set_theme_replaces_markdown_theme_instead_of_stacking(self) -> None:
        original_console = renderer.console
        original_stdout_console = renderer._stdout_console
        renderer.console, _ = _capture_console()
        renderer._stdout_console, _ = _capture_console()
        try:
            baseline = len(renderer.console._theme_stack._entries)
            set_theme(CliTheme.load("dawn"))
            after_first = len(renderer.console._theme_stack._entries)
            set_theme(CliTheme.load("midnight"))
            after_second = len(renderer.console._theme_stack._entries)
        finally:
            renderer.console = original_console
            renderer._stdout_console = original_stdout_console
        assert after_first == baseline + 1
        assert after_second == baseline + 1


class TestRefreshAliases:
    def test_refresh_aliases_updates_gold(self) -> None:
        renderer._theme = CliTheme.load("dawn")
        _refresh_aliases()
        assert renderer.GOLD == CliTheme.load("dawn").accent

    def test_refresh_aliases_updates_muted(self) -> None:
        renderer._theme = CliTheme.load("high-contrast")
        _refresh_aliases()
        assert renderer.MUTED == CliTheme.load("high-contrast").muted

    def test_refresh_aliases_updates_chrome(self) -> None:
        renderer._theme = CliTheme.load("accessible")
        _refresh_aliases()
        assert renderer.CHROME == CliTheme.load("accessible").chrome

    def test_refresh_aliases_all_fields_in_sync(self) -> None:
        theme = CliTheme.load("dawn")
        renderer._theme = theme
        _refresh_aliases()
        assert renderer.GOLD == theme.accent
        assert renderer.SLATE == theme.secondary
        assert renderer.BLUE == theme.logo_blue
        assert renderer.MUTED == theme.muted
        assert renderer.CHROME == theme.chrome
        assert renderer.ERROR_RED == theme.error


# ---------------------------------------------------------------------------
# Dawn theme accent
# ---------------------------------------------------------------------------


class TestDawnThemeAccent:
    def test_dawn_accent_value(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer._theme.accent == "#B8860B"

    def test_dawn_error_value(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer._theme.error == "#DC2626"

    def test_dawn_warning_value(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        assert renderer._theme.warning == "#D97706"


# ---------------------------------------------------------------------------
# Midnight theme aliases
# ---------------------------------------------------------------------------


class TestMidnightAliases:
    def test_midnight_gold_alias(self) -> None:
        set_theme(CliTheme.load("midnight"))
        assert renderer.GOLD == "#C5A059"

    def test_midnight_muted_alias(self) -> None:
        set_theme(CliTheme.load("midnight"))
        assert renderer.MUTED == "#8b8b8b"

    def test_midnight_chrome_alias(self) -> None:
        set_theme(CliTheme.load("midnight"))
        assert renderer.CHROME == "#6b7280"

    def test_midnight_error_alias(self) -> None:
        set_theme(CliTheme.load("midnight"))
        assert renderer.ERROR_RED == "#CD6B6B"


# ---------------------------------------------------------------------------
# render_error() uses theme error color
# ---------------------------------------------------------------------------


class TestRenderErrorTheme:
    def test_render_error_uses_theme_error_color(self) -> None:
        """render_error() must embed the theme's error color, not 'red'."""
        dawn = CliTheme.load("dawn")
        set_theme(dawn)

        buf = io.StringIO()
        # Patch the module-level console so render_error() writes to our buffer
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_error("something went wrong")
        finally:
            renderer.console = original_console

        output = buf.getvalue()
        # Rich converts #RRGGBB to ANSI 24-bit sequences; check for the RGB portion
        assert _hex_to_ansi_rgb(dawn.error) in output
        assert "something went wrong" in _plain(output)

    def test_render_error_output_contains_error_label(self) -> None:
        set_theme(CliTheme.load("midnight"))

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_error("disk full")
        finally:
            renderer.console = original_console

        plain = _plain(buf.getvalue())
        assert "Error:" in plain
        assert "disk full" in plain

    def test_render_error_midnight_error_color(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_error("connection refused")
        finally:
            renderer.console = original_console

        assert _hex_to_ansi_rgb(midnight.error) in buf.getvalue()

    def test_render_error_no_hardcoded_red(self) -> None:
        """'red' must not appear as a standalone markup tag in error output."""
        set_theme(CliTheme.load("midnight"))

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_error("boom")
        finally:
            renderer.console = original_console

        # Rich markup uses [red] or [red bold]; neither should be present
        # when the theme drives the color via a hex value
        raw = buf.getvalue()
        assert "[red]" not in raw
        assert "[red bold]" not in raw


# ---------------------------------------------------------------------------
# render_warning() uses theme warning color
# ---------------------------------------------------------------------------


class TestRenderWarningTheme:
    def test_render_warning_uses_theme_warning_color(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_warning("low memory")
        finally:
            renderer.console = original_console

        output = buf.getvalue()
        assert _hex_to_ansi_rgb(dawn.warning) in output
        assert "low memory" in _plain(output)

    def test_render_warning_output_contains_warning_label(self) -> None:
        set_theme(CliTheme.load("midnight"))

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_warning("approaching token limit")
        finally:
            renderer.console = original_console

        plain = _plain(buf.getvalue())
        assert "Warning:" in plain
        assert "approaching token limit" in plain

    def test_render_warning_no_hardcoded_yellow(self) -> None:
        set_theme(CliTheme.load("midnight"))

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_warning("something")
        finally:
            renderer.console = original_console

        raw = buf.getvalue()
        assert "[yellow]" not in raw
        assert "[yellow bold]" not in raw

    def test_render_warning_accessible_theme(self) -> None:
        accessible = CliTheme.load("accessible")
        set_theme(accessible)

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_warning("cvd safe warning")
        finally:
            renderer.console = original_console

        assert _hex_to_ansi_rgb(accessible.warning) in buf.getvalue()


# ---------------------------------------------------------------------------
# _build_thinking_text() uses theme colors
# ---------------------------------------------------------------------------


class TestBuildThinkingTextTheme:
    def test_thinking_text_uses_accent_color(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)
        # elapsed must be >= _REPL_THINKING_REVEAL_DELAY (2s) to produce output (#1052)
        text = _build_thinking_text(2.5)
        # The accent ANSI code should appear in the output
        assert midnight.ansi_fg("accent") in text

    def test_thinking_text_elapsed_uses_chrome_color(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)
        # Use elapsed > 3.0 to get the timer line (#1052 raised calm window to 3s)
        text = _build_thinking_text(4.0)
        assert midnight.ansi_fg("chrome") in text

    def test_thinking_text_error_uses_error_color(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)
        text = _build_thinking_text(1.0, error_msg="timeout")
        assert midnight.ansi_fg("error") in text

    def test_thinking_text_dawn_accent(self) -> None:
        dawn = CliTheme.load("dawn")
        set_theme(dawn)
        text = _build_thinking_text(2.5)
        assert dawn.ansi_fg("accent") in text

    def test_thinking_text_high_contrast_accent(self) -> None:
        hc = CliTheme.load("high-contrast")
        set_theme(hc)
        text = _build_thinking_text(2.5)
        assert hc.ansi_fg("accent") in text

    def test_thinking_text_muted_in_cancel(self) -> None:
        midnight = CliTheme.load("midnight")
        set_theme(midnight)
        text = _build_thinking_text(1.0, cancel_msg="cancelled")
        assert midnight.ansi_fg("muted") in text


# ---------------------------------------------------------------------------
# All built-in themes round-trip through set_theme()
# ---------------------------------------------------------------------------


class TestAllBuiltInThemes:
    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_set_theme_returns_correct_theme_object(self, theme_name: str) -> None:
        expected = CliTheme.load(theme_name)
        set_theme(expected)
        assert renderer._theme is expected

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_aliases_match_theme_after_set(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)
        assert renderer.GOLD == theme.accent
        assert renderer.MUTED == theme.muted
        assert renderer.CHROME == theme.chrome
        assert renderer.ERROR_RED == theme.error

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_render_error_uses_theme_error_for_each_builtin(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_error("test error")
        finally:
            renderer.console = original_console

        assert _hex_to_ansi_rgb(theme.error) in buf.getvalue()

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_render_warning_uses_theme_warning_for_each_builtin(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)

        buf = io.StringIO()
        original_console = renderer.console
        renderer.console = Console(file=buf, force_terminal=True, color_system="truecolor", width=120, highlight=False)
        try:
            render_warning("test warning")
        finally:
            renderer.console = original_console

        assert _hex_to_ansi_rgb(theme.warning) in buf.getvalue()

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_theme_accent_is_non_empty(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        assert theme.accent != ""

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_theme_error_is_non_empty(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        assert theme.error != ""

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_theme_warning_is_non_empty(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        assert theme.warning != ""


# ---------------------------------------------------------------------------
# Busy toolbar visual snapshot tests (#1134)
# ---------------------------------------------------------------------------


class TestBusyToolbarTheme:
    """Verify that busy-state toolbar segments use the correct theme colors."""

    @pytest.fixture(autouse=True)
    def restore_midnight_theme(self) -> None:
        yield
        set_theme(CliTheme.load("midnight"))

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_toolbar_busy_thinking_text_present(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)
        busy = BusyStatus(thinking_text="Thinking... 6s")
        result = format_status_toolbar(model="gpt-4o", busy_status=busy)
        text = "".join(t[1] for t in result)
        assert "Thinking... 6s" in text

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_toolbar_busy_thinking_uses_warn_style(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)
        busy = BusyStatus(thinking_text="Thinking... 6s")
        result = format_status_toolbar(busy_status=busy)
        thinking_parts = [t for t in result if "Thinking" in t[1]]
        assert thinking_parts
        assert thinking_parts[0][0] == "class:bottom-toolbar.tokens-warn"

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_toolbar_busy_cancel_hint_uses_dim_style(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)
        busy = BusyStatus(thinking_text="Thinking... 10s", show_cancel_hint=True)
        result = format_status_toolbar(busy_status=busy)
        cancel_parts = [t for t in result if "esc to cancel" in t[1]]
        assert cancel_parts
        assert cancel_parts[0][0] == "class:bottom-toolbar.dim"

    @pytest.mark.parametrize("theme_name", list(_BUILTIN_THEMES.keys()))
    def test_toolbar_busy_tool_label_uses_model_style(self, theme_name: str) -> None:
        theme = CliTheme.load(theme_name)
        set_theme(theme)
        busy = BusyStatus(thinking_text="", tool_label="grep")
        result = format_status_toolbar(busy_status=busy)
        tool_parts = [t for t in result if "grep" in t[1]]
        assert tool_parts
        assert tool_parts[0][0] == "class:bottom-toolbar.model"



"""Tests for the conversation rewind service and endpoint."""

from __future__ import annotations

import sqlite3
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import AsyncMock, patch

import pytest

from anteroom.db import _FTS_SCHEMA, _FTS_TRIGGERS, _SCHEMA, ThreadSafeConnection
from anteroom.models import RewindRequest
from anteroom.routers.conversations import rewind_conversation
from anteroom.services.rewind import collect_file_paths
from anteroom.services.rewind import rewind_conversation as rewind_service
from anteroom.services.storage import (
    create_conversation,
    create_message,
    create_tool_call,
    list_messages,
    list_tool_calls,
)


@pytest.fixture()
def db() -> ThreadSafeConnection:
    conn = sqlite3.connect(":memory:", check_same_thread=False)
    conn.row_factory = sqlite3.Row
    conn.execute("PRAGMA foreign_keys=ON")
    conn.executescript(_SCHEMA)
    try:
        conn.executescript(_FTS_SCHEMA)
        conn.executescript(_FTS_TRIGGERS)
    except sqlite3.OperationalError:
        pass
    conn.commit()
    return ThreadSafeConnection(conn)


def _make_request(db: ThreadSafeConnection, data_dir: Path | None = None) -> SimpleNamespace:
    """Build a fake Request with app.state for the endpoint."""
    config = SimpleNamespace(app=SimpleNamespace(data_dir=data_dir))
    state = SimpleNamespace(db=db, config=config)
    app = SimpleNamespace(state=state)
    request = SimpleNamespace(app=app, query_params={})
    return request


class TestRewindService:
    """Tests for the shared rewind service layer."""

    async def test_rewind_deletes_messages(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        create_message(db, conv["id"], "assistant", "msg1")
        create_message(db, conv["id"], "user", "msg2")
        create_message(db, conv["id"], "assistant", "msg3")

        result = await rewind_service(db, conv["id"], to_position=1, data_dir=tmp_path)

        assert result.deleted_messages == 2
        assert result.reverted_files == []
        assert result.skipped_files == []
        msgs = list_messages(db, conv["id"])
        assert len(msgs) == 2
        assert [m["content"] for m in msgs] == ["msg0", "msg1"]

    async def test_undo_files_reverts_git(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "wrote a file")
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/test.py", "content": "hello"})

        mock_proc = AsyncMock()
        mock_proc.returncode = 0
        mock_proc.communicate = AsyncMock(return_value=(b"", b""))

        with (
            patch("anteroom.services.rewind.check_git_repo", return_value=True),
            patch("asyncio.create_subprocess_exec", return_value=mock_proc) as mock_exec,
        ):
            result = await rewind_service(db, conv["id"], to_position=0, undo_files=True, data_dir=tmp_path)

        assert result.deleted_messages == 1
        assert len(result.reverted_files) == 1
        assert "/tmp/test.py" in result.reverted_files[0]
        assert result.skipped_files == []
        mock_exec.assert_called_once()

    async def test_undo_files_skips_on_git_failure(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "wrote a file")
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/test.py", "content": "hello"})

        mock_proc = AsyncMock()
        mock_proc.returncode = 1
        mock_proc.communicate = AsyncMock(return_value=(b"", b"error: pathspec did not match"))

        with (
            patch("anteroom.services.rewind.check_git_repo", return_value=True),
            patch("asyncio.create_subprocess_exec", return_value=mock_proc),
        ):
            result = await rewind_service(db, conv["id"], to_position=0, undo_files=True, data_dir=tmp_path)

        assert result.deleted_messages == 1
        assert result.reverted_files == []
        assert len(result.skipped_files) == 1
        assert "pathspec" in result.skipped_files[0]

    async def test_undo_files_skips_non_git(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "wrote a file")
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/test.py", "content": "hello"})

        with patch("anteroom.services.rewind.check_git_repo", return_value=False):
            result = await rewind_service(db, conv["id"], to_position=0, undo_files=True, data_dir=tmp_path)

        assert result.deleted_messages == 1
        assert result.reverted_files == []
        assert len(result.skipped_files) == 1
        assert "not a git repo" in result.skipped_files[0]

    async def test_no_file_tools(self, db, tmp_path):
        """Rewind with undo_files=True but no file-modifying tool calls."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "used a tool")
        create_tool_call(db, msg1["id"], "bash", "builtin", {"command": "ls"})

        result = await rewind_service(db, conv["id"], to_position=0, undo_files=True, data_dir=tmp_path)

        assert result.deleted_messages == 1
        assert result.reverted_files == []
        assert result.skipped_files == []

    async def test_deduplicates_file_paths(self, db, tmp_path):
        """Same file modified by multiple tool calls should only be reverted once."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "edited twice")
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/test.py", "content": "v1"})
        create_tool_call(
            db,
            msg1["id"],
            "edit_file",
            "builtin",
            {"path": "/tmp/test.py", "old_string": "v1", "new_string": "v2"},
        )

        mock_proc = AsyncMock()
        mock_proc.returncode = 0
        mock_proc.communicate = AsyncMock(return_value=(b"", b""))

        with (
            patch("anteroom.services.rewind.check_git_repo", return_value=True),
            patch("asyncio.create_subprocess_exec", return_value=mock_proc) as mock_exec,
        ):
            result = await rewind_service(db, conv["id"], to_position=0, undo_files=True, data_dir=tmp_path)

        assert len(result.reverted_files) == 1
        mock_exec.assert_called_once()

    async def test_cascades_tool_calls(self, db, tmp_path):
        """Tool calls on deleted messages should be cleaned up."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "msg1")
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/f.py", "content": "x"})

        await rewind_service(db, conv["id"], to_position=0, data_dir=tmp_path)

        tcs = list_tool_calls(db, msg1["id"])
        assert len(tcs) == 0

    def test_collect_file_paths(self, db):
        """collect_file_paths extracts paths from write_file and edit_file only."""
        conv = create_conversation(db, title="Test")
        msg = create_message(db, conv["id"], "assistant", "tools")
        create_tool_call(db, msg["id"], "write_file", "builtin", {"path": "/a.py", "content": "x"})
        create_tool_call(db, msg["id"], "edit_file", "builtin", {"path": "/b.py", "old_string": "x", "new_string": "y"})
        create_tool_call(db, msg["id"], "bash", "builtin", {"command": "echo hi"})
        create_tool_call(db, msg["id"], "read_file", "builtin", {"path": "/c.py"})

        paths = collect_file_paths(db, [msg["id"]])
        assert paths == {"/a.py", "/b.py"}


class TestRewindEndpoint:
    """Tests for the HTTP endpoint wrapper around the rewind service."""

    async def test_rewind_endpoint(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        create_message(db, conv["id"], "assistant", "msg1")

        request = _make_request(db, tmp_path)
        body = RewindRequest(to_position=0, undo_files=False)

        result = await rewind_conversation(conv["id"], body, request)

        assert result.deleted_messages == 1
        msgs = list_messages(db, conv["id"])
        assert len(msgs) == 1

    async def test_rewind_invalid_position(self, db, tmp_path):
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")

        request = _make_request(db, tmp_path)
        body = RewindRequest(to_position=99, undo_files=False)

        with pytest.raises(Exception, match="Invalid position"):
            await rewind_conversation(conv["id"], body, request)

    async def test_rewind_nonexistent_conversation(self, db, tmp_path):
        request = _make_request(db, tmp_path)
        body = RewindRequest(to_position=0, undo_files=False)

        with pytest.raises(Exception, match="Conversation not found"):
            await rewind_conversation("00000000-0000-0000-0000-000000000000", body, request)


class TestRewindSpecificPosition:
    """Additional edge cases for rewind to specific positions."""

    async def test_rewind_to_specific_message(self, db, tmp_path):
        """Rewind to a specific middle position preserves messages up to that point."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        create_message(db, conv["id"], "assistant", "msg1")
        create_message(db, conv["id"], "user", "msg2")
        create_message(db, conv["id"], "assistant", "msg3")
        create_message(db, conv["id"], "user", "msg4")

        result = await rewind_service(db, conv["id"], to_position=2, data_dir=tmp_path)

        assert result.deleted_messages == 2
        msgs = list_messages(db, conv["id"])
        assert len(msgs) == 3
        assert [m["content"] for m in msgs] == ["msg0", "msg1", "msg2"]

    async def test_rewind_no_messages_to_delete(self, db, tmp_path):
        """Rewind to position beyond all messages deletes nothing."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")

        result = await rewind_service(db, conv["id"], to_position=5, data_dir=tmp_path)

        assert result.deleted_messages == 0
        msgs = list_messages(db, conv["id"])
        assert len(msgs) == 1

    async def test_rewind_past_tool_calls(self, db, tmp_path):
        """Rewind deletes tool calls associated with rewound messages."""
        conv = create_conversation(db, title="Test")
        create_message(db, conv["id"], "user", "msg0")
        msg1 = create_message(db, conv["id"], "assistant", "used tools")
        create_tool_call(db, msg1["id"], "bash", "builtin", {"command": "ls"})
        create_tool_call(db, msg1["id"], "write_file", "builtin", {"path": "/tmp/f.py", "content": "x"})
        create_message(db, conv["id"], "user", "msg2")

        result = await rewind_service(db, conv["id"], to_position=0, data_dir=tmp_path)

        assert result.deleted_messages == 2
        tcs = list_tool_calls(db, msg1["id"])
        assert len(tcs) == 0



"""Tests for encryption at rest functionality."""

from __future__ import annotations

import sqlite3
from pathlib import Path
from unittest.mock import patch

import pytest

from anteroom.services.encryption import (
    HKDF_INFO,
    HKDF_KEY_LENGTH,
    HKDF_SALT,
    derive_db_key,
    is_sqlcipher_available,
    verify_encryption,
)

# A stable test PEM key for deterministic key derivation tests
_TEST_PEM = """-----BEGIN PRIVATE KEY-----
MC4CAQAwBQYDK2VwBCIEIKFwEEeil3TLxMALbFxB7pVqpOSyIosFX5mP5+g6mFYy
-----END PRIVATE KEY-----"""


class TestDeriveDbKey:
    def test_derives_32_byte_key(self) -> None:
        key = derive_db_key(_TEST_PEM)
        assert isinstance(key, bytes)
        assert len(key) == HKDF_KEY_LENGTH

    def test_deterministic(self) -> None:
        key1 = derive_db_key(_TEST_PEM)
        key2 = derive_db_key(_TEST_PEM)
        assert key1 == key2

    def test_different_keys_for_different_pems(self) -> None:
        key1 = derive_db_key(_TEST_PEM)
        key2 = derive_db_key("-----BEGIN PRIVATE KEY-----\nDIFFERENT\n-----END PRIVATE KEY-----")
        assert key1 != key2

    def test_empty_pem_raises(self) -> None:
        with pytest.raises(ValueError, match="no identity private key"):
            derive_db_key("")

    def test_uses_correct_hkdf_params(self) -> None:
        """Verify the HKDF parameters are what we expect for key derivation."""
        assert HKDF_SALT == b"anteroom-db-encryption-salt-v1"
        assert HKDF_INFO == b"anteroom-db-encryption-v1"
        assert HKDF_KEY_LENGTH == 32


class TestIsSqlcipherAvailable:
    def test_returns_false_when_import_fails(self) -> None:
        import sys

        # Remove any cached module so the import is retried
        saved = sys.modules.pop("sqlcipher3", None)
        try:
            with patch.dict("sys.modules", {"sqlcipher3": None}):
                # Reload the encryption module to pick up the patched import
                # But is_sqlcipher_available does a deferred import, so just call it
                # with sqlcipher3 set to None (triggers ImportError on attribute access)
                assert is_sqlcipher_available() is False
        finally:
            if saved is not None:
                sys.modules["sqlcipher3"] = saved

    def test_returns_true_when_installed(self) -> None:
        import sys
        from unittest.mock import MagicMock

        mock_mod = MagicMock()
        saved = sys.modules.get("sqlcipher3")
        sys.modules["sqlcipher3"] = mock_mod
        try:
            assert is_sqlcipher_available() is True
        finally:
            if saved is not None:
                sys.modules["sqlcipher3"] = saved
            else:
                sys.modules.pop("sqlcipher3", None)


class TestVerifyEncryption:
    def test_plaintext_db_returns_false(self, tmp_path: Path) -> None:
        db_path = tmp_path / "plain.db"
        conn = sqlite3.connect(str(db_path))
        conn.execute("CREATE TABLE test (id INTEGER)")
        conn.commit()
        conn.close()

        assert verify_encryption(db_path) is False

    def test_nonexistent_db_returns_false(self, tmp_path: Path) -> None:
        assert verify_encryption(tmp_path / "nope.db") is False

    def test_corrupted_file_returns_true(self, tmp_path: Path) -> None:
        db_path = tmp_path / "encrypted.db"
        db_path.write_bytes(b"\x00" * 4096 + b"random encrypted data")
        assert verify_encryption(db_path) is True


class TestOpenEncryptedDb:
    def test_open_calls_sqlcipher_with_pragmas(self) -> None:
        """open_encrypted_db calls sqlcipher3.connect and sets PRAGMA key."""
        import sys
        from unittest.mock import MagicMock

        from anteroom.services.encryption import open_encrypted_db

        mock_conn = MagicMock()
        mock_conn.execute.return_value.fetchone.return_value = (5,)
        mock_sqlcipher = MagicMock()
        mock_sqlcipher.connect.return_value = mock_conn

        saved = sys.modules.get("sqlcipher3")
        sys.modules["sqlcipher3"] = mock_sqlcipher
        try:
            key = derive_db_key(_TEST_PEM)
            conn = open_encrypted_db(Path("/tmp/test.db"), key)
            assert conn is mock_conn
            mock_sqlcipher.connect.assert_called_once_with("/tmp/test.db")
            # Verify PRAGMA key was set
            calls = [str(c) for c in mock_conn.execute.call_args_list]
            assert any("PRAGMA key" in c for c in calls)
            assert any("PRAGMA cipher_page_size" in c for c in calls)
        finally:
            if saved is not None:
                sys.modules["sqlcipher3"] = saved
            else:
                sys.modules.pop("sqlcipher3", None)

    def test_open_raises_on_wrong_key(self) -> None:
        """open_encrypted_db raises ValueError when DB can't be read with key."""
        import sys
        from unittest.mock import MagicMock

        from anteroom.services.encryption import open_encrypted_db

        mock_conn = MagicMock()
        # First two execute calls (PRAGMA key, cipher_page_size) succeed
        # Third call (SELECT count) raises
        call_count = 0

        def _side_effect(sql, *args):
            nonlocal call_count
            call_count += 1
            if "SELECT count" in sql:
                raise Exception("file is not a database")
            return MagicMock()

        mock_conn.execute.side_effect = _side_effect
        mock_sqlcipher = MagicMock()
        mock_sqlcipher.connect.return_value = mock_conn

        saved = sys.modules.get("sqlcipher3")
        sys.modules["sqlcipher3"] = mock_sqlcipher
        try:
            key = derive_db_key(_TEST_PEM)
            with pytest.raises(ValueError, match="Failed to open encrypted database"):
                open_encrypted_db(Path("/tmp/test.db"), key)
            mock_conn.close.assert_called_once()
        finally:
            if saved is not None:
                sys.modules["sqlcipher3"] = saved
            else:
                sys.modules.pop("sqlcipher3", None)


class TestMigration:
    def test_migration_nonexistent_db_raises(self, tmp_path: Path) -> None:
        from anteroom.services.encryption import migrate_plaintext_to_encrypted

        key = derive_db_key(_TEST_PEM)
        with pytest.raises(FileNotFoundError):
            migrate_plaintext_to_encrypted(tmp_path / "nope.db", key)

    def test_migration_already_encrypted_raises(self, tmp_path: Path) -> None:
        from anteroom.services.encryption import migrate_plaintext_to_encrypted

        db_path = tmp_path / "encrypted.db"
        db_path.write_bytes(b"\x00" * 4096 + b"looks encrypted")

        key = derive_db_key(_TEST_PEM)
        with pytest.raises(ValueError, match="already appears to be encrypted"):
            migrate_plaintext_to_encrypted(db_path, key)

    def test_migration_happy_path_with_mock(self, tmp_path: Path) -> None:
        """Test migration flow: backup, export, verify, replace."""
        import os
        import sys
        from unittest.mock import MagicMock

        from anteroom.services.encryption import migrate_plaintext_to_encrypted

        # Create a real plaintext DB
        db_path = tmp_path / "chat.db"
        conn = sqlite3.connect(str(db_path))
        conn.execute("CREATE TABLE test (id INTEGER, name TEXT)")
        conn.execute("INSERT INTO test VALUES (1, 'hello')")
        conn.commit()
        conn.close()

        key = derive_db_key(_TEST_PEM)

        # Mock sqlcipher3 to simulate the migration
        mock_src_conn = MagicMock()
        mock_verify_conn = MagicMock()
        mock_verify_conn.execute.return_value.fetchone.return_value = (3,)  # non-zero table count

        connect_calls = []

        def _mock_connect(path):
            connect_calls.append(path)
            if len(connect_calls) == 1:
                return mock_src_conn
            return mock_verify_conn

        mock_sqlcipher = MagicMock()
        mock_sqlcipher.connect.side_effect = _mock_connect

        saved = sys.modules.get("sqlcipher3")
        sys.modules["sqlcipher3"] = mock_sqlcipher
        try:
            # Write a fake encrypted DB to the temp path that migration will create
            # We need to patch tempfile.mkstemp to control the temp file
            with patch("anteroom.services.encryption.tempfile.mkstemp") as mock_mkstemp:
                tmp_encrypted = tmp_path / "tmp_encrypted.db"
                tmp_encrypted.write_bytes(b"fake encrypted")
                fd = os.open(str(tmp_encrypted), os.O_RDWR)
                mock_mkstemp.return_value = (fd, str(tmp_encrypted))

                backup_path = migrate_plaintext_to_encrypted(db_path, key)

            # Verify backup was created
            assert backup_path.exists()
            assert backup_path.suffix.endswith("-plaintext")

            # Verify sqlcipher_export was called
            export_calls = [str(c) for c in mock_src_conn.execute.call_args_list]
            assert any("sqlcipher_export" in c for c in export_calls)
        finally:
            if saved is not None:
                sys.modules["sqlcipher3"] = saved
            else:
                sys.modules.pop("sqlcipher3", None)


class TestStorageConfig:
    def test_default_values(self) -> None:
        from anteroom.config import StorageConfig

        cfg = StorageConfig()
        assert cfg.retention_days == 0
        assert cfg.retention_check_interval == 3600
        assert cfg.purge_attachments is True
        assert cfg.purge_embeddings is True
        assert cfg.encrypt_at_rest is False
        assert cfg.encryption_kdf == "hkdf-sha256"

    def test_clamps_check_interval(self) -> None:
        from anteroom.config import StorageConfig

        cfg = StorageConfig(retention_check_interval=10)
        assert cfg.retention_check_interval == 60  # clamped to minimum

    def test_custom_values(self) -> None:
        from anteroom.config import StorageConfig

        cfg = StorageConfig(
            retention_days=90,
            retention_check_interval=7200,
            purge_attachments=False,
            encrypt_at_rest=True,
        )
        assert cfg.retention_days == 90
        assert cfg.retention_check_interval == 7200
        assert cfg.purge_attachments is False
        assert cfg.encrypt_at_rest is True


class TestConfigValidation:
    def test_storage_section_accepted(self) -> None:
        from anteroom.services.config_validator import validate_config

        raw = {
            "ai": {"base_url": "http://localhost:11434/v1"},
            "storage": {
                "retention_days": 90,
                "retention_check_interval": 3600,
                "encrypt_at_rest": False,
            },
        }
        result = validate_config(raw)
        assert result.is_valid

    def test_unknown_storage_key_warns(self) -> None:
        from anteroom.services.config_validator import validate_config

        raw = {
            "ai": {"base_url": "http://localhost:11434/v1"},
            "storage": {"unknown_field": True},
        }
        result = validate_config(raw)
        warnings = [e for e in result.errors if e.severity == "warning"]
        assert any("unknown_field" in str(w) for w in warnings)

    def test_retention_days_range_validation(self) -> None:
        from anteroom.services.config_validator import validate_config

        raw = {
            "ai": {"base_url": "http://localhost:11434/v1"},
            "storage": {"retention_days": -1},
        }
        result = validate_config(raw)
        warnings = [e for e in result.errors if "retention_days" in e.path]
        assert len(warnings) > 0


class TestDbInitWithEncryption:
    def test_init_db_without_encryption(self, tmp_path: Path) -> None:
        from anteroom.db import init_db

        db = init_db(tmp_path / "test.db")
        assert db is not None
        db.close()

    def test_init_db_encryption_key_none_uses_sqlite(self, tmp_path: Path) -> None:
        from anteroom.db import init_db

        db = init_db(tmp_path / "test.db", encryption_key=None)
        assert db is not None
        db.close()



"""Tests for workflow cost and token tracking (#963).

Token tracking is always-on (not gated on budget). Cost estimation
uses model_costs rates passed to WorkflowEngine.
"""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest

from anteroom.config import WorkflowBudgetConfig, WorkflowConfig
from anteroom.db import init_db
from anteroom.services.workflow_engine import WorkflowEngine, load_definition
from anteroom.services.workflow_runners import RunnerResult, create_default_registry
from anteroom.services.workflow_storage import get_workflow_run

# ---------------------------------------------------------------------------
# Test YAML (domain-neutral)
# ---------------------------------------------------------------------------

ONE_STEP_WORKFLOW = """\
kind: workflow
id: test_cost
version: 0.1.0
inputs: {}
policies: {}
steps:
  - id: step_one
    type: runner
    runner: shell
    command: "echo hello"
"""

TWO_STEP_WORKFLOW = """\
kind: workflow
id: test_cost_two
version: 0.1.0
inputs: {}
policies: {}
steps:
  - id: step_one
    type: runner
    runner: shell
    command: "echo one"
  - id: step_two
    type: runner
    runner: shell
    command: "echo two"
"""

BUDGET_WORKFLOW = """\
kind: workflow
id: test_cost_budget
version: 0.1.0
inputs: {}
policies:
  budget:
    max_steps: 10
    max_tokens: 100000
steps:
  - id: step_one
    type: runner
    runner: shell
    command: "echo one"
  - id: step_two
    type: runner
    runner: shell
    command: "echo two"
"""


@pytest.fixture()
def db():
    with tempfile.TemporaryDirectory() as td:
        conn = init_db(Path(td) / "test.db")
        yield conn


@pytest.fixture()
def config():
    return WorkflowConfig()


@pytest.fixture()
def registry():
    return create_default_registry()


def _make_engine(
    db: Any,
    config: WorkflowConfig,
    registry: Any,
    *,
    model_costs: dict[str, dict[str, float]] | None = None,
) -> WorkflowEngine:
    return WorkflowEngine(db, config, registry, model_costs=model_costs)


async def _token_runner(**kwargs: Any) -> RunnerResult:
    return RunnerResult(
        status="success",
        summary="done",
        duration_ms=10,
        artifacts={
            "total_tokens": 100,
            "prompt_tokens": 70,
            "completion_tokens": 30,
            "model": "gpt-4o",
        },
    )


async def _zero_token_runner(**kwargs: Any) -> RunnerResult:
    return RunnerResult(status="success", summary="done", duration_ms=10)


# ===========================================================================
# Token tracking without budget
# ===========================================================================


class TestTokenTrackingWithoutBudget:
    @pytest.mark.asyncio
    async def test_token_tracking_without_budget(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Token tracking works even when no budget is configured."""
        engine = _make_engine(db, config, registry)
        defn = load_definition(ONE_STEP_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-1")

        assert run["status"] == "completed"
        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"] is not None
        assert stored["budget_usage"]["total_tokens"] == 100
        assert stored["budget_usage"]["prompt_tokens"] == 70
        assert stored["budget_usage"]["completion_tokens"] == 30
        assert stored["budget_usage"]["steps_completed"] == 1

    @pytest.mark.asyncio
    async def test_token_tracking_with_budget(self, db: Any, registry: Any) -> None:
        """Token tracking works alongside budget enforcement."""
        cfg = WorkflowConfig(budget=WorkflowBudgetConfig(max_steps=10, max_tokens=100000))
        engine = _make_engine(db, cfg, registry)
        defn = load_definition(BUDGET_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-2")

        assert run["status"] == "completed"
        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"]["total_tokens"] == 200  # 2 steps * 100
        assert stored["budget"] is not None
        assert stored["budget"]["max_tokens"] == 100000


# ===========================================================================
# Model name in artifacts
# ===========================================================================


class TestModelNameInArtifacts:
    @pytest.mark.asyncio
    async def test_model_name_in_artifacts(self) -> None:
        """Agent runner includes model name in result artifacts."""
        from anteroom.services.agent_loop import AgentEvent
        from anteroom.services.workflow_runners import execute_agent_runner

        mock_events = [
            AgentEvent(kind="token", data={"content": "hello"}),
            AgentEvent(kind="usage", data={"prompt_tokens": 50, "completion_tokens": 25}),
        ]

        async def mock_agent_loop(**kwargs: Any) -> Any:
            for event in mock_events:
                yield event

        from unittest.mock import AsyncMock

        mock_ai = AsyncMock()
        mock_executor = AsyncMock(return_value={"status": "ok"})

        with patch("anteroom.services.agent_loop.run_agent_loop", side_effect=mock_agent_loop):
            result = await execute_agent_runner(
                prompt="test",
                ai_service=mock_ai,
                tool_executor=mock_executor,
                tools_openai=[],
                model="gpt-4o",
            )

        assert result.artifacts["model"] == "gpt-4o"

    @pytest.mark.asyncio
    async def test_model_name_defaults_to_empty(self) -> None:
        """Agent runner defaults model to empty string when not provided."""
        from anteroom.services.agent_loop import AgentEvent
        from anteroom.services.workflow_runners import execute_agent_runner

        mock_events = [AgentEvent(kind="token", data={"content": "hi"})]

        async def mock_agent_loop(**kwargs: Any) -> Any:
            for event in mock_events:
                yield event

        from unittest.mock import AsyncMock

        mock_ai = AsyncMock()
        mock_executor = AsyncMock(return_value={"status": "ok"})

        with patch("anteroom.services.agent_loop.run_agent_loop", side_effect=mock_agent_loop):
            result = await execute_agent_runner(
                prompt="test",
                ai_service=mock_ai,
                tool_executor=mock_executor,
                tools_openai=[],
            )

        assert result.artifacts["model"] == ""


# ===========================================================================
# Cost estimation
# ===========================================================================


class TestCostEstimation:
    def test_cost_estimation_known_model(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Known model with rates produces correct cost."""
        model_costs = {"gpt-4o": {"input": 2.50, "output": 10.00}}
        engine = _make_engine(db, config, registry, model_costs=model_costs)
        artifacts = {"model": "gpt-4o", "prompt_tokens": 1_000_000, "completion_tokens": 500_000}
        cost = engine._estimate_step_cost(artifacts)
        expected = (1_000_000 / 1_000_000) * 2.50 + (500_000 / 1_000_000) * 10.00
        assert cost == pytest.approx(expected)

    def test_cost_estimation_unknown_model(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Unknown model returns 0 cost."""
        model_costs = {"gpt-4o": {"input": 2.50, "output": 10.00}}
        engine = _make_engine(db, config, registry, model_costs=model_costs)
        artifacts = {"model": "unknown-model", "prompt_tokens": 1000, "completion_tokens": 500}
        assert engine._estimate_step_cost(artifacts) == 0.0

    def test_cost_estimation_no_model_costs(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Engine without model_costs returns 0 cost."""
        engine = _make_engine(db, config, registry)
        artifacts = {"model": "gpt-4o", "prompt_tokens": 1000, "completion_tokens": 500}
        assert engine._estimate_step_cost(artifacts) == 0.0

    def test_cost_estimation_missing_model_key(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Artifacts without model key returns 0 cost."""
        model_costs = {"gpt-4o": {"input": 2.50, "output": 10.00}}
        engine = _make_engine(db, config, registry, model_costs=model_costs)
        artifacts = {"prompt_tokens": 1000, "completion_tokens": 500}
        assert engine._estimate_step_cost(artifacts) == 0.0


# ===========================================================================
# Mixed runners and accumulation
# ===========================================================================


class TestTokenAccumulation:
    @pytest.mark.asyncio
    async def test_mixed_runner_tokens(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Shell steps contribute 0 tokens; agent steps contribute actual tokens."""
        engine = _make_engine(db, config, registry)
        defn = load_definition(TWO_STEP_WORKFLOW)

        call_count = 0

        async def _mixed_runner(**kwargs: Any) -> RunnerResult:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return RunnerResult(status="success", summary="done", duration_ms=10)
            return RunnerResult(
                status="success",
                summary="done",
                duration_ms=10,
                artifacts={"total_tokens": 200, "prompt_tokens": 120, "completion_tokens": 80},
            )

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_mixed_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-mix")

        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"]["total_tokens"] == 200
        assert stored["budget_usage"]["prompt_tokens"] == 120
        assert stored["budget_usage"]["completion_tokens"] == 80

    @pytest.mark.asyncio
    async def test_token_accumulation_across_steps(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """Two steps with tokens sum correctly."""
        engine = _make_engine(db, config, registry)
        defn = load_definition(TWO_STEP_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-accum")

        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"]["total_tokens"] == 200
        assert stored["budget_usage"]["prompt_tokens"] == 140
        assert stored["budget_usage"]["completion_tokens"] == 60
        assert stored["budget_usage"]["steps_completed"] == 2

    @pytest.mark.asyncio
    async def test_budget_usage_always_persisted(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """budget_usage_json is never None after a step completes."""
        engine = _make_engine(db, config, registry)
        defn = load_definition(ONE_STEP_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_zero_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-persist")

        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"] is not None
        assert stored["budget_usage"]["steps_completed"] == 1

    @pytest.mark.asyncio
    async def test_estimated_cost_in_budget_usage(self, db: Any, config: WorkflowConfig, registry: Any) -> None:
        """estimated_cost_usd is present in stored budget_usage."""
        model_costs = {"gpt-4o": {"input": 2.50, "output": 10.00}}
        engine = _make_engine(db, config, registry, model_costs=model_costs)
        defn = load_definition(ONE_STEP_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-est")

        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        usage = stored["budget_usage"]
        assert "estimated_cost_usd" in usage
        expected = (70 / 1_000_000) * 2.50 + (30 / 1_000_000) * 10.00
        assert usage["estimated_cost_usd"] == pytest.approx(expected)

    @pytest.mark.asyncio
    async def test_estimated_cost_zero_without_model_costs(
        self, db: Any, config: WorkflowConfig, registry: Any
    ) -> None:
        """Without model_costs, estimated_cost_usd is 0."""
        engine = _make_engine(db, config, registry)
        defn = load_definition(ONE_STEP_WORKFLOW)

        with patch("anteroom.services.workflow_runners.execute_opaque_runner", side_effect=_token_runner):
            run = await engine.start_run(defn, target_kind="test", target_ref="cost-zero")

        stored = get_workflow_run(db, run["id"])
        assert stored is not None
        assert stored["budget_usage"]["estimated_cost_usd"] == 0.0



"""Unit tests for canvas CRUD endpoints in the conversations router."""

from __future__ import annotations

import uuid
from unittest.mock import MagicMock, patch

from fastapi import FastAPI
from fastapi.testclient import TestClient

from anteroom.routers.conversations import router

_CANVAS = {
    "id": str(uuid.uuid4()),
    "conversation_id": str(uuid.uuid4()),
    "title": "Test Canvas",
    "content": "# Hello",
    "language": None,
    "version": 1,
    "created_at": "2024-01-01T00:00:00",
    "updated_at": "2024-01-01T00:00:00",
    "user_id": None,
    "user_display_name": None,
}

_CONV = {
    "id": _CANVAS["conversation_id"],
    "title": "Test",
    "type": "chat",
    "created_at": "2024-01-01",
    "updated_at": "2024-01-01",
}


def _make_app() -> FastAPI:
    app = FastAPI()
    app.include_router(router, prefix="/api")
    mock_db = MagicMock()
    mock_db_manager = MagicMock()
    mock_db_manager.get.return_value = mock_db
    app.state.db = mock_db
    app.state.db_manager = mock_db_manager
    mock_config = MagicMock()
    mock_config.identity = None
    mock_config.app.data_dir = MagicMock()
    app.state.config = mock_config
    return app


class TestCreateCanvasEndpoint:
    def test_create_canvas_success(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = None
            mock_storage.create_canvas.return_value = _CANVAS
            client = TestClient(app)
            resp = client.post(
                f"/api/conversations/{conv_id}/canvas",
                json={"title": "Test Canvas", "content": "# Hello"},
            )
        assert resp.status_code == 201
        assert resp.json()["title"] == "Test Canvas"

    def test_create_canvas_409_duplicate(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            client = TestClient(app)
            resp = client.post(
                f"/api/conversations/{conv_id}/canvas",
                json={"title": "Dup", "content": "C"},
            )
        assert resp.status_code == 409

    def test_create_canvas_404_no_conversation(self) -> None:
        app = _make_app()
        fake_id = str(uuid.uuid4())
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = None
            client = TestClient(app)
            resp = client.post(
                f"/api/conversations/{fake_id}/canvas",
                json={"title": "T", "content": "C"},
            )
        assert resp.status_code == 404

    def test_create_canvas_invalid_uuid(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/api/conversations/not-a-uuid/canvas",
            json={"title": "T", "content": "C"},
        )
        assert resp.status_code == 400

    def test_create_canvas_title_too_long(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        client = TestClient(app)
        resp = client.post(
            f"/api/conversations/{conv_id}/canvas",
            json={"title": "x" * 201, "content": "C"},
        )
        assert resp.status_code == 422


class TestGetCanvasEndpoint:
    def test_get_canvas_success(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            client = TestClient(app)
            resp = client.get(f"/api/conversations/{conv_id}/canvas")
        assert resp.status_code == 200
        assert resp.json()["content"] == "# Hello"

    def test_get_canvas_404_no_canvas(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = None
            client = TestClient(app)
            resp = client.get(f"/api/conversations/{conv_id}/canvas")
        assert resp.status_code == 404

    def test_get_canvas_404_no_conversation(self) -> None:
        app = _make_app()
        fake_id = str(uuid.uuid4())
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = None
            client = TestClient(app)
            resp = client.get(f"/api/conversations/{fake_id}/canvas")
        assert resp.status_code == 404

    def test_get_canvas_invalid_uuid(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.get("/api/conversations/not-a-uuid/canvas")
        assert resp.status_code == 400


class TestUpdateCanvasEndpoint:
    def test_update_canvas_success(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        updated = {**_CANVAS, "content": "updated", "version": 2}
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            mock_storage.update_canvas.return_value = updated
            client = TestClient(app)
            resp = client.patch(
                f"/api/conversations/{conv_id}/canvas",
                json={"content": "updated"},
            )
        assert resp.status_code == 200
        assert resp.json()["version"] == 2

    def test_update_canvas_404_no_canvas(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = None
            client = TestClient(app)
            resp = client.patch(
                f"/api/conversations/{conv_id}/canvas",
                json={"content": "data"},
            )
        assert resp.status_code == 404

    def test_update_canvas_404_no_conversation(self) -> None:
        app = _make_app()
        fake_id = str(uuid.uuid4())
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = None
            client = TestClient(app)
            resp = client.patch(
                f"/api/conversations/{fake_id}/canvas",
                json={"content": "data"},
            )
        assert resp.status_code == 404

    def test_update_canvas_invalid_uuid(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.patch(
            "/api/conversations/not-a-uuid/canvas",
            json={"content": "data"},
        )
        assert resp.status_code == 400

    def test_update_canvas_title_only(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        updated = {**_CANVAS, "title": "New Title Only", "version": 2}
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            mock_storage.update_canvas.return_value = updated
            client = TestClient(app)
            resp = client.patch(
                f"/api/conversations/{conv_id}/canvas",
                json={"title": "New Title Only"},
            )
        assert resp.status_code == 200
        assert resp.json()["title"] == "New Title Only"
        assert resp.json()["content"] == _CANVAS["content"]

    def test_update_canvas_returns_none(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            mock_storage.update_canvas.return_value = None
            client = TestClient(app)
            resp = client.patch(
                f"/api/conversations/{conv_id}/canvas",
                json={"content": "data"},
            )
        assert resp.status_code == 404


class TestDeleteCanvasEndpoint:
    def test_delete_canvas_success(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = _CANVAS
            mock_storage.delete_canvas.return_value = True
            client = TestClient(app)
            resp = client.delete(f"/api/conversations/{conv_id}/canvas")
        assert resp.status_code == 204

    def test_delete_canvas_404_no_canvas(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = _CONV
            mock_storage.get_canvas_for_conversation.return_value = None
            client = TestClient(app)
            resp = client.delete(f"/api/conversations/{conv_id}/canvas")
        assert resp.status_code == 404

    def test_delete_canvas_404_no_conversation(self) -> None:
        app = _make_app()
        fake_id = str(uuid.uuid4())
        with patch("anteroom.routers.conversations.storage") as mock_storage:
            mock_storage.get_conversation.return_value = None
            client = TestClient(app)
            resp = client.delete(f"/api/conversations/{fake_id}/canvas")
        assert resp.status_code == 404

    def test_delete_canvas_invalid_uuid(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.delete("/api/conversations/not-a-uuid/canvas")
        assert resp.status_code == 400


class TestCanvasContentTypeEnforcement:
    """Canvas endpoints — Content-Type enforcement."""

    def test_create_canvas_rejects_non_json(self) -> None:
        """Non-JSON body rejected: Pydantic fires 422 before _require_json can return 415."""
        app = _make_app()
        conv_id = _CONV["id"]
        client = TestClient(app)
        resp = client.post(
            f"/api/conversations/{conv_id}/canvas",
            content="title=Test",
            headers={"content-type": "text/plain"},
        )
        assert resp.status_code == 422

    def test_update_canvas_rejects_non_json(self) -> None:
        """Non-JSON body rejected: Pydantic fires 422 before _require_json can return 415."""
        app = _make_app()
        conv_id = _CONV["id"]
        client = TestClient(app)
        resp = client.patch(
            f"/api/conversations/{conv_id}/canvas",
            content="content=data",
            headers={"content-type": "text/plain"},
        )
        assert resp.status_code == 422

    def test_create_canvas_rejects_empty_title(self) -> None:
        app = _make_app()
        conv_id = _CONV["id"]
        client = TestClient(app)
        resp = client.post(
            f"/api/conversations/{conv_id}/canvas",
            json={"title": "", "content": "C"},
        )
        assert resp.status_code == 422



"""Tests for paste detection and input collapsing in the CLI REPL."""

from __future__ import annotations

import os
import time
from unittest.mock import MagicMock, patch

import pytest
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document

from anteroom.cli.repl import _MAX_PASTE_DISPLAY_LINES, _PASTE_THRESHOLD, _collapse_long_input, _is_paste

_TERM_SIZE = os.terminal_size((80, 24))


class TestIsPaste:
    def test_recent_change_is_paste(self) -> None:
        """Buffer change < 50ms ago → paste."""
        just_now = time.monotonic()
        assert _is_paste(just_now) is True

    def test_old_change_is_not_paste(self) -> None:
        """Buffer change > 50ms ago → normal typing."""
        long_ago = time.monotonic() - 0.2  # 200ms ago
        assert _is_paste(long_ago) is False

    def test_exactly_at_threshold_is_not_paste(self) -> None:
        """Buffer change exactly at threshold boundary → not paste."""
        at_threshold = time.monotonic() - _PASTE_THRESHOLD - 0.001
        assert _is_paste(at_threshold) is False

    def test_custom_threshold(self) -> None:
        """Custom threshold is respected."""
        recent = time.monotonic() - 0.02  # 20ms ago
        assert _is_paste(recent, threshold=0.01) is False
        assert _is_paste(recent, threshold=0.05) is True

    def test_zero_timestamp_is_not_paste(self) -> None:
        """Initial state (0.0) should never count as paste."""
        assert _is_paste(0.0) is False


class TestPasteKeybindingBehavior:
    """Simulate the Enter key handler logic with mocked events."""

    def _make_event(self) -> MagicMock:
        event = MagicMock()
        event.current_buffer = MagicMock()
        return event

    def test_enter_during_paste_inserts_newline(self) -> None:
        """When Enter arrives during a paste burst, insert newline."""
        event = self._make_event()
        last_change = [time.monotonic()]  # just changed

        # Simulate the Enter handler logic
        if _is_paste(last_change[0]):
            event.current_buffer.insert_text("\n")
        else:
            event.current_buffer.validate_and_handle()

        event.current_buffer.insert_text.assert_called_once_with("\n")
        event.current_buffer.validate_and_handle.assert_not_called()

    def test_enter_after_delay_submits(self) -> None:
        """When Enter arrives well after last change, submit."""
        event = self._make_event()
        last_change = [time.monotonic() - 0.2]  # 200ms ago

        if _is_paste(last_change[0]):
            event.current_buffer.insert_text("\n")
        else:
            event.current_buffer.validate_and_handle()

        event.current_buffer.validate_and_handle.assert_called_once()
        event.current_buffer.insert_text.assert_not_called()

    def test_rapid_multiple_newlines_all_insert(self) -> None:
        """Simulate pasting 3 lines: each newline should insert, not submit."""
        event = self._make_event()
        submitted = False

        for _ in range(3):
            last_change = time.monotonic()  # simulate rapid buffer change
            if _is_paste(last_change):
                event.current_buffer.insert_text("\n")
            else:
                submitted = True

        assert not submitted
        assert event.current_buffer.insert_text.call_count == 3

    def test_enter_after_paste_settles_submits(self) -> None:
        """After paste is done and user waits, Enter submits."""
        event = self._make_event()

        # Simulate: paste happened, then user reviews for 500ms, then presses Enter
        last_change = [time.monotonic() - 0.5]

        if _is_paste(last_change[0]):
            event.current_buffer.insert_text("\n")
        else:
            event.current_buffer.validate_and_handle()

        event.current_buffer.validate_and_handle.assert_called_once()


class TestCollapseLongInput:
    """Tests for _collapse_long_input() output rendering."""

    def _make_lines(self, count: int) -> str:
        return "\n".join(f"line {i}" for i in range(count))

    def test_short_input_no_output(self) -> None:
        """Input with <= _MAX_PASTE_DISPLAY_LINES produces no output."""
        short = self._make_lines(_MAX_PASTE_DISPLAY_LINES)
        with (
            patch("anteroom.cli.repl.renderer") as mock_renderer,
            patch("anteroom.cli.repl.sys") as mock_sys,
        ):
            mock_sys.stdout.isatty.return_value = True
            _collapse_long_input(short)
            mock_renderer.console.print.assert_not_called()
            mock_renderer._stdout.write.assert_not_called()

    def test_long_input_collapses(self) -> None:
        """Input with many lines writes collapsed output via renderer."""
        long_input = self._make_lines(20)
        with (
            patch("anteroom.cli.repl.renderer") as mock_renderer,
            patch("anteroom.cli.repl.sys") as mock_sys,
            patch("anteroom.cli.repl.shutil") as mock_shutil,
        ):
            mock_sys.stdout.isatty.return_value = True

            mock_shutil.get_terminal_size.return_value = _TERM_SIZE
            _collapse_long_input(long_input)

            # Cursor movement via _stdout (real fd)
            assert mock_renderer._stdout.write.call_count == 1
            cursor_call = mock_renderer._stdout.write.call_args[0][0]
            assert "\033[" in cursor_call  # ANSI cursor movement
            assert "\033[J" in cursor_call  # clear to end

            # 3 shown lines + 1 hidden count = 4 console.print calls
            assert mock_renderer.console.print.call_count == 4

            # First call has the prompt character
            first_call = mock_renderer.console.print.call_args_list[0][0][0]
            assert "❯" in first_call

            # Last call has the hidden line count
            last_call = mock_renderer.console.print.call_args_list[3][0][0]
            assert "17 more lines" in last_call

            mock_renderer._stdout.flush.assert_called_once()

    def test_user_content_escaped(self) -> None:
        """Pasted content with Rich markup is escaped in output."""
        lines = ["[bold red]danger[/]"] + [f"line {i}" for i in range(19)]
        raw = "\n".join(lines)
        with (
            patch("anteroom.cli.repl.renderer") as mock_renderer,
            patch("anteroom.cli.repl.sys") as mock_sys,
            patch("anteroom.cli.repl.shutil") as mock_shutil,
        ):
            mock_sys.stdout.isatty.return_value = True

            mock_shutil.get_terminal_size.return_value = _TERM_SIZE
            _collapse_long_input(raw)

            first_call = mock_renderer.console.print.call_args_list[0][0][0]
            # The raw markup should be escaped — Rich escape() turns [ into \[
            assert "[bold red]" not in first_call or "\\[bold red]" in first_call

    def test_not_a_tty_no_output(self) -> None:
        """When stdout is not a TTY, function returns early."""
        long_input = self._make_lines(20)
        with (
            patch("anteroom.cli.repl.renderer") as mock_renderer,
            patch("anteroom.cli.repl.sys") as mock_sys,
        ):
            mock_sys.stdout.isatty.return_value = False
            _collapse_long_input(long_input)
            mock_renderer.console.print.assert_not_called()
            mock_renderer._stdout.write.assert_not_called()


class TestAcceptCompletionBehavior:
    """Simulate the _accept_completion() logic with mocked buffers."""

    def _make_buf(self, *, has_completion: bool = False) -> MagicMock:
        buf = MagicMock()
        if has_completion:
            completion = MagicMock()
            buf.complete_state = MagicMock()
            buf.complete_state.current_completion = completion
        else:
            buf.complete_state = None
        buf.completer = MagicMock()
        return buf

    @staticmethod
    def _accept_completion(buf: MagicMock) -> bool:
        """Reproduce the _accept_completion logic from repl.py."""
        if buf.complete_state and buf.complete_state.current_completion:
            saved_completer = buf.completer
            buf.completer = None
            try:
                buf.apply_completion(buf.complete_state.current_completion)
            finally:
                buf.completer = saved_completer
            return True
        return False

    def test_accepts_when_completion_menu_open(self) -> None:
        """When the completion menu has a selection, accept it."""
        buf = self._make_buf(has_completion=True)
        result = self._accept_completion(buf)
        assert result is True
        buf.apply_completion.assert_called_once()

    def test_returns_false_when_no_completion(self) -> None:
        """When no completion menu is open, return False."""
        buf = self._make_buf(has_completion=False)
        result = self._accept_completion(buf)
        assert result is False
        buf.apply_completion.assert_not_called()

    def test_temporarily_nulls_completer(self) -> None:
        """Completer is temporarily set to None during apply to prevent _Retry."""
        buf = self._make_buf(has_completion=True)
        original_completer = buf.completer

        # Track completer state during apply_completion
        completer_during_apply = []

        def capture_completer(*args: object) -> None:
            completer_during_apply.append(buf.completer)

        buf.apply_completion.side_effect = capture_completer
        self._accept_completion(buf)

        # During apply_completion, completer should have been None
        assert completer_during_apply == [None]
        # After, completer should be restored
        assert buf.completer is original_completer

    def test_completer_restored_on_exception(self) -> None:
        """#617-12: Completer must be restored even if apply_completion raises."""
        buf = self._make_buf(has_completion=True)
        original_completer = buf.completer
        buf.apply_completion.side_effect = RuntimeError("boom")

        with pytest.raises(RuntimeError, match="boom"):
            self._accept_completion(buf)

        # Completer must still be restored despite the exception
        assert buf.completer is original_completer

    def test_enter_with_completion_does_not_submit(self) -> None:
        """Enter key accepts completion instead of submitting when menu is open."""
        buf = self._make_buf(has_completion=True)
        last_change = [time.monotonic() - 0.5]  # not a paste

        # Simulate Enter handler logic
        if self._accept_completion(buf):
            submitted = False
        elif _is_paste(last_change[0]):
            submitted = False
        else:
            submitted = True

        assert submitted is False
        buf.apply_completion.assert_called_once()
        buf.validate_and_handle.assert_not_called()

    def test_enter_without_completion_submits(self) -> None:
        """Enter key submits when no completion menu is open."""
        buf = self._make_buf(has_completion=False)
        last_change = [time.monotonic() - 0.5]  # not a paste

        if self._accept_completion(buf):
            pass
        elif _is_paste(last_change[0]):
            buf.insert_text("\n")
        else:
            buf.validate_and_handle()

        buf.validate_and_handle.assert_called_once()

    def test_tab_starts_completion_when_no_menu(self) -> None:
        """Tab triggers start_completion when no menu is open."""
        buf = self._make_buf(has_completion=False)
        buf.complete_state = None

        # Simulate Tab handler logic
        if buf.complete_state:
            if not self._accept_completion(buf):
                buf.complete_next()
        else:
            buf.start_completion()

        buf.start_completion.assert_called_once()

    def test_tab_accepts_current_completion(self) -> None:
        """Tab accepts the current completion when menu is open."""
        buf = self._make_buf(has_completion=True)

        # Simulate Tab handler logic
        if buf.complete_state:
            if not self._accept_completion(buf):
                buf.complete_next()
        else:
            buf.start_completion()

        buf.apply_completion.assert_called_once()
        buf.complete_next.assert_not_called()


class TestShouldAutoComplete:
    """Test the auto-complete trigger predicate logic."""

    @staticmethod
    def _should_auto_complete(text: str) -> bool:
        """Reproduce the _should_auto_complete logic from repl.py."""
        stripped = text.lstrip()
        if stripped.startswith("/") and " " not in stripped:
            return True
        if "@" in (text.split()[-1] if text.split() else ""):
            return True
        return False

    def test_slash_command_triggers(self) -> None:
        assert self._should_auto_complete("/hel") is True

    def test_slash_command_with_leading_spaces(self) -> None:
        assert self._should_auto_complete("  /hel") is True

    def test_slash_command_with_space_after_does_not_trigger(self) -> None:
        assert self._should_auto_complete("/help ") is False

    def test_at_sign_triggers(self) -> None:
        assert self._should_auto_complete("read @src/") is True

    def test_at_sign_in_last_word_triggers(self) -> None:
        assert self._should_auto_complete("@file.py") is True

    def test_plain_text_does_not_trigger(self) -> None:
        assert self._should_auto_complete("hello world") is False

    def test_empty_text_does_not_trigger(self) -> None:
        assert self._should_auto_complete("") is False

    def test_slash_only_triggers(self) -> None:
        assert self._should_auto_complete("/") is True

    def test_full_command_with_args_does_not_trigger(self) -> None:
        assert self._should_auto_complete("/resume my-slug") is False


class TestAnteroomCompleter:
    """Tests for the AnteroomCompleter class defined in repl.py."""

    def _make_completer(self) -> MagicMock:
        """Create a mock that simulates AnteroomCompleter.get_completions logic."""
        commands = ["help", "resume", "model", "quit"]
        skill_names = ["commit", "deploy"]
        skill_descriptions = {"commit": "Create a commit", "deploy": "Deploy to PyPI"}
        command_descriptions = {
            "help": "show help",
            "resume": "resume conversation",
            "model": "change model",
            "quit": "exit",
        }

        completer = MagicMock()

        def get_completions(document: Document, complete_event: MagicMock) -> list[Completion]:
            text = document.text_before_cursor
            word = document.get_word_before_cursor(WORD=True)
            results = []

            stripped = text.lstrip()
            if stripped.startswith("/") and " " not in stripped:
                prefix = word.lstrip("/")
                for cmd in commands:
                    if cmd.startswith(prefix):
                        meta = command_descriptions.get(cmd, "")
                        results.append(Completion(f"/{cmd} ", start_position=-len(word), display_meta=meta))
                for sname in skill_names:
                    if sname.startswith(prefix):
                        desc = skill_descriptions.get(sname, "skill")
                        results.append(Completion(f"/{sname} ", start_position=-len(word), display_meta=desc))
            return results

        completer.get_completions = get_completions
        return completer

    def test_slash_yields_matching_commands(self) -> None:
        completer = self._make_completer()
        doc = Document("/he", cursor_position=3)
        completions = list(completer.get_completions(doc, MagicMock()))
        texts = [c.text for c in completions]
        assert "/help " in texts

    def test_slash_yields_skills(self) -> None:
        completer = self._make_completer()
        doc = Document("/co", cursor_position=3)
        completions = list(completer.get_completions(doc, MagicMock()))
        texts = [c.text for c in completions]
        assert "/commit " in texts

    def test_skill_has_display_meta(self) -> None:
        completer = self._make_completer()
        doc = Document("/de", cursor_position=3)
        completions = list(completer.get_completions(doc, MagicMock()))
        deploy_completions = [c for c in completions if "/deploy" in c.text]
        assert len(deploy_completions) == 1
        assert "Deploy to PyPI" in str(deploy_completions[0].display_meta)

    def test_command_has_display_meta(self) -> None:
        completer = self._make_completer()
        doc = Document("/he", cursor_position=3)
        completions = list(completer.get_completions(doc, MagicMock()))
        help_completions = [c for c in completions if "/help" in c.text]
        assert len(help_completions) == 1
        assert "show help" in str(help_completions[0].display_meta)

    def test_completions_have_trailing_space(self) -> None:
        completer = self._make_completer()
        doc = Document("/he", cursor_position=3)
        completions = list(completer.get_completions(doc, MagicMock()))
        for c in completions:
            assert c.text.endswith(" ")

    def test_no_completions_after_space(self) -> None:
        completer = self._make_completer()
        doc = Document("/help foo", cursor_position=9)
        completions = list(completer.get_completions(doc, MagicMock()))
        assert completions == []

    def test_empty_slash_yields_all(self) -> None:
        completer = self._make_completer()
        doc = Document("/", cursor_position=1)
        completions = list(completer.get_completions(doc, MagicMock()))
        # Should yield all commands + all skills
        assert len(completions) == 6  # 4 commands + 2 skills


class TestOnAcceptBehavior:
    """Test the _on_accept buffer handler logic."""

    def test_signals_main_input_when_no_sub_prompt(self) -> None:
        """Accept stashes text and signals _input_ready."""
        accepted_text = [""]
        input_ready = MagicMock()
        sub_prompt_event = None

        buf = MagicMock()
        buf.text = "hello world"

        # Simulate _on_accept logic
        accepted_text[0] = buf.text
        if sub_prompt_event is not None:
            sub_prompt_event.set()
        else:
            input_ready.set()

        assert accepted_text[0] == "hello world"
        input_ready.set.assert_called_once()

    def test_signals_sub_prompt_when_active(self) -> None:
        """Accept signals sub-prompt event instead of main input."""
        accepted_text = [""]
        input_ready = MagicMock()
        sub_prompt_event = MagicMock()

        buf = MagicMock()
        buf.text = "sub response"

        # Simulate _on_accept logic
        accepted_text[0] = buf.text
        if sub_prompt_event is not None:
            sub_prompt_event.set()
        else:
            input_ready.set()

        assert accepted_text[0] == "sub response"
        sub_prompt_event.set.assert_called_once()
        input_ready.set.assert_not_called()

    def test_returns_true_to_keep_text(self) -> None:
        """_on_accept returns True to keep text in buffer (cleared after reading)."""
        # The return value True tells prompt_toolkit to keep text in the buffer
        # We test this by simulating the handler
        result = True  # _on_accept always returns True
        assert result is True
